from __future__ import annotations
from .cast import cast_list, cast_tuple, cast_int, cast_date
from .context import PROXY_CONTEXT, REDIRECT_CONTEXT, GCP_CONTEXT
from .date import now, get_date, cast_datetime_format, get_date_range
from .gcloud import Account, NumericiseIgnore, BigQuerySchema, SchemaSequence
from .gcloud import fetch_gcloud_authorization, update_gspread, read_gspread
from .gcloud import read_gbq, to_gbq, validate_upsert_key, IDTokenCredentials
from .logs import CustomLogger, dumps_map, unraw
from .types import _KT, _VT, ClassInstance, Context, ContextMapper, IndexLabel, Keyword, DateFormat, DateQuery, Timedelta, TypeHint
from .types import RenameDict, JsonData, RedirectData, LogMessage, Records, TabularData, Data, Unit
from .types import is_array, is_records, init_origin
from .map import unique, exists_one, fill_array, is_same_length, align_array
from .map import kloc, apply_dict, chain_dict, drop_dict
from .map import cloc, apply_df, merge_drop, convert_data, filter_data, chain_exists, data_empty
from .parse import parse_cookies, parse_origin

from abc import ABCMeta, abstractmethod
from urllib.parse import unquote, urlparse
import asyncio
import aiohttp
import functools
import requests
import time

from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Type, Union
from collections import defaultdict
import datetime as dt
import json
import logging
import pandas as pd

from tqdm.auto import tqdm
from itertools import product
import base64
import random
import sys
import traceback


EXAMPLE_URL = "https://example.com"

HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "ko-KR,ko;q=0.9,en-US;q=0.8,en;q=0.7",
    "Connection": "keep-alive",
    "sec-ch-ua": '"Chromium";v="116", "Not)A;Brand";v="24", "Google Chrome";v="116"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": '"Windows"',
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-origin",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
}

MAX_ASYNC_TASK_LIMIT = 100
MAX_REDIRECT_LIMIT = 10

GATHER_MSG = lambda which, where: f"Collecting {which} from {where}."
REDIRECT_MSG = lambda operation: f"{operation} operation is redirecting."
INVALID_REDIRECT_LOG_MSG = "Please verify that the both results and errors are in redierct returns."
LOGIN_REQUIRED_MSG = lambda where=str(): f"{where} login information is required.".strip()
LOGIN_INVALID_MSG = lambda where=str(): f"{where} login information is not valid.".strip()
DEPENDENCY_HAS_NO_NAME_MSG = "Dependency has no operation name. Please define operation name."


encrypt = lambda s=str(), count=0, *args: encrypt(
    base64.b64encode(str(s).encode("utf-8")).decode("utf-8"), count-1) if count else s
decrypt = lambda s=str(), count=0, *args: decrypt(
    base64.b64decode(str(s).encode("utf-8")).decode("utf-8"), count-1) if count else s


def get_cookies(session, url=None) -> str:
    if isinstance(session, aiohttp.ClientSession):
        return "; ".join([str(key)+"="+str(value) for key,value in session.cookie_jar.filter_cookies(url).items()])
    elif isinstance(session, requests.Session):
        return "; ".join([str(key)+"="+str(value) for key,value in dict(session.cookies).items()])


def get_headers(authority=str(), referer=str(), cookies=str(), host=str(),
                origin: Optional[Union[bool,str]]=False, **kwargs) -> Dict:
    headers = HEADERS.copy()
    if authority: headers["Authority"] = urlparse(authority).hostname
    if referer: headers["referer"] = referer
    if host: headers["Host"] = urlparse(host).hostname
    if origin: headers["Origin"] = parse_origin(origin if isinstance(origin, str) else (authority if authority else host))
    if cookies: headers["Cookie"] = cookies
    return dict(headers, **kwargs)


###################################################################
############################### Base ##############################
###################################################################

class CustomDict(dict):
    def __init__(self, **kwargs):
        super().__init__(kwargs)

    def copy(self, __instance: Optional[ClassInstance]=None) -> Any:
        if __instance: return __instance.__class__(**self.__dict__)
        else: return self.__class__(**self.__dict__)

    def get(self, __key: _KT, default=None, cast=False) -> Union[Any,Dict,List,str]:
        return kloc(self.__dict__, __key, default, cast)

    def update(self, __m: Optional[Dict]=dict(), inplace=True, **kwargs) -> Union[bool,Dict]:
        if not inplace: self = self.copy()
        for key, value in dict(kwargs, **__m).items():
            setattr(self, key, value)
        super().update(self.__dict__)
        return exists_one(inplace, self)

    def __getitem__(self, __keys: _KT) -> _VT:
        if is_array(__keys):
            return [super().__getitem__(__key) for __key in __keys]
        else:
            return super().__getitem__(__keys)

    def __setitem__(self, __key: _KT, __value: _VT):
        setattr(self, __key, __value)
        super().__setitem__(__key, __value)


###################################################################
############################# Spiders #############################
###################################################################

class Spider(CustomDict):
    __metaclass__ = ABCMeta
    asyncio = False
    operation = "spider"
    where = "urls"
    which = "data"
    contextArgs = list()
    contextQuery = list()
    contextUnit = 1
    contextPeriod = str()
    returnType = "records"
    errors = defaultdict(list)
    maxLimit = 1
    message = str()
    rename = dict()

    def __init__(self, operation=str(), where=str(), which=str(), filter: IndexLabel=list(), filterContext: Optional[ContextMapper]=None,
                contextArgs: List[_KT]=list(), contextQuery: List[_KT]=list(), contextUnit: Unit=1,
                contextPeriod: Timedelta=str(), startDate: Optional[DateFormat]=None, endDate: Optional[DateFormat]=None,
                returnType: Optional[TypeHint]=str(), logName=str(), logLevel: Union[str,int]="WARN", logFile: Optional[str]=str(),
                logErrors=False, logJson=False, errorArgs: IndexLabel=tuple(), errorKwargs: IndexLabel=tuple(), errors: Dict=dict(),
                delay: Union[float,int,Tuple[int]]=1., numTasks=100, maxLimit=1, progress=True, debug=False,
                message=str(), rename: RenameDict=dict(), apiRedirect=False, reidrectUnit: Unit=1, redirectErrors=False,
                localSave=False, extraSave=False, dependencies: Tuple[Type]=tuple(), self_var=True, **context):
        self.operation = self.operation
        self.initTime = now()
        self.filter = filter
        self.contextArgs = contextArgs if contextArgs else self.contextArgs
        self.contextQuery = contextQuery if contextQuery else self.contextQuery
        self.contextUnit = contextUnit if contextUnit else self.contextUnit
        self.contextPeriod = contextPeriod if contextPeriod else self.contextPeriod
        self.logName = logName if logName else self.operation
        self.logLevel = int(logLevel) if str(logLevel).isdigit() else logging.getLevelName(str(logLevel).upper())
        self.logFile = logFile
        self.logJson = bool(logFile)
        self.logger = CustomLogger(name=self.logName, level=self.logLevel, file=self.logFile)
        self.logErrors = logErrors
        self.errorArgs = errorArgs
        self.errorKwargs = errorKwargs
        self.returnType = returnType if returnType else self.returnType
        self.delay = delay
        self.numTasks = min(cast_int(numTasks, default=self.maxLimit), self.maxLimit)
        self.progress = progress
        self.debug = debug
        self.message = message if message else self.message
        self.rename = rename if rename else self.rename
        self.apiRedirect = apiRedirect
        self.reidrectUnit = reidrectUnit
        self.redirectErrors = redirectErrors
        self.localSave = localSave
        self.extraSave = extraSave
        self.set_context(filterContext=filterContext, **context)
        self.set_date(startDate=startDate, endDate=endDate, contextPeriod=contextPeriod, **context)
        self.set_query(**context)

    def set_context(self, filterContext: Optional[ContextMapper]=None, **context):
        self.update(filterContext(**context) if isinstance(filterContext, Callable) else context)

    def set_date(self, startDate: Optional[DateFormat]=None, endDate: Optional[DateFormat]=None,
                contextPeriod: Timedelta=str(), **context):
        if contextPeriod or (startDate != None) or (endDate != None):
            startDate, endDate = get_date(startDate, default=1), get_date(endDate, default=1)
            self.startDate = min(startDate, endDate)
            self.endDate = max(startDate, endDate)

    ###################################################################
    ######################### Session Managers ########################
    ###################################################################

    def requests_task(func):
        @functools.wraps(func)
        def wrapper(self: Spider, *args, self_var=True, **kwargs):
            if self_var: kwargs = dict(self.__dict__, **kwargs)
            data = func(self, *args, **kwargs)
            self.upload_data(data, **GCP_CONTEXT(**kwargs))
            return data
        return wrapper

    def requests_session(func):
        @functools.wraps(func)
        def wrapper(self: Spider, *args, self_var=True, **kwargs):
            if self_var: kwargs = dict(self.__dict__, **kwargs)
            with requests.Session() as session:
                data = func(self, *args, session=session, **kwargs)
            time.sleep(.25)
            self.upload_data(data, **GCP_CONTEXT(**kwargs))
            return data
        return wrapper

    def log_errors(func):
        @functools.wraps(func)
        def wrapper(self: Spider, *args, **kwargs):
            try: return func(self, *args, **kwargs)
            except KeyboardInterrupt as interrupt:
                raise interrupt
            except Exception as exception:
                if self.logErrors: self.log_arguments(args, kwargs)
                if self.debug: raise exception
                name = f"{func.__name__}({self.__class__.__name__})"
                self.logger.error(log_exception(name, *args, json=self.logJson, **kwargs))
                return init_origin(func)
        return wrapper

    def log_arguments(self, args, kwargs):
        for idx, key in enumerate(self.errorArgs):
            if idx < len(args) and key: self.errors[key].append(args[idx])
        for key in self.errorKwargs:
            if key in kwargs: self.errors[key].append(kwargs[key])

    def requests_limit(func):
        @functools.wraps(func)
        def wrapper(self: Spider, *args, **kwargs):
            if self.delay: time.sleep(self.get_delay())
            return func(self, *args, **kwargs)
        return wrapper

    def get_delay(self, tsUnit="ms", **kwargs) -> Union[float,int]:
        if isinstance(self.delay, (float,int)): return self.delay
        elif isinstance(self.delay, Tuple):
            random.randrange(*self.delay[:2])/(1000 if tsUnit == "ms" else 1)
        else: return 0.

    ###################################################################
    ######################### Gather Requests #########################
    ###################################################################

    @abstractmethod
    @requests_session
    def crawl(self, **context) -> Data:
        return self.gather(**self.map_context(**context))

    def map_context(self, __unique=True, **context) -> Context:
        return {key: (unique(*value) if __unique else value)
                if is_array(value) else value for key, value in context.items()}

    def gather(self, *args, message=str(), progress=None, filter: IndexLabel=list(),
                returnType: Optional[TypeHint]=None, **context) -> Data:
        message = message if message else GATHER_MSG(self.which, self.where)
        hide_bar = not (progress if isinstance(progress, bool) else self.progress)
        query, context = self.split_context(*args, **context)
        data = [self.fetch(**__q, filter=filter, **context) for __q in tqdm(query, desc=message, disable=hide_bar)]
        return self.map_reduce(data, filter=filter, returnType=returnType, **context)

    def map_reduce(self, data: List, filter=list(), returnType: Optional[TypeHint]=None, **kwargs) -> Data:
        return filter_data(chain_exists(data), filter=filter, returnType=returnType)

    ###################################################################
    ########################## Split Context ##########################
    ###################################################################

    def split_context(self, *args, contextArgs: List[_KT]=list(), contextQuery: List[_KT]=list(),
                    contextUnit: Unit=1, contextPeriod: Timedelta=str(), **context) -> Tuple[List[Context],Context]:
        if args and contextArgs:
            contextQuery = contextArgs + contextQuery
            context = dict({key: args[idx] for idx, key in enumerate(cast_tuple(contextArgs)[:len(args)])}, **context)
        query, context = self.from_context(contextQuery, contextUnit, **context)
        if contextPeriod:
            period, context = self.from_period(contextPeriod=contextPeriod, **context)
            query = self.product_context(query, period)
        return query, context

    def from_context(self, contextQuery: List[_KT], contextUnit: Unit=1, **context) -> Tuple[List[Context],Context]:
        if not contextQuery: return list(), context
        query, context = kloc(context, contextQuery, default="pass"), drop_dict(context, contextQuery, inplace=False)
        contextUnit = cast_list(contextUnit)
        if any(map(lambda x: x>1, contextUnit)): query = self.group_context(contextQuery, contextUnit, **query)
        else: query = [dict(zip(query.keys(), values)) for values in product(*map(cast_tuple, query.values()))]
        return query, context

    def group_context(self, contextQuery: List[_KT], contextUnit: Unit=1, **context) -> List[Context]:
        query = apply_dict(context, apply=cast_list, all_keys=True)
        keys, unit = query.keys(), fill_array(contextUnit, count=len(contextQuery), value=1)
        combinations = product(*[range(0, len(query[key]), unit[i]) for i, key in enumerate(keys)])
        return [{key: query[key][index:index+unit[i]] for i, (key, index) in enumerate(zip(keys, indices))}
                for indices in combinations]

    def from_period(self, startDate: Optional[dt.date]=None, endDate: Optional[dt.date]=None,
                    contextPeriod: Timedelta="D", **context) -> Tuple[List[DateQuery],Context]:
        startDate = startDate if isinstance(startDate, dt.date) else cast_date(startDate, default=self.startDate)
        endDate = endDate if isinstance(endDate, dt.date) else cast_date(endDate, default=self.endDate)
        date_range = get_date_range(startDate, endDate, interval=contextPeriod)
        if (contextPeriod in ("D",1)) or (str(contextPeriod).startswith("1 day")):
            period = [dict(date=date) for date in date_range]
            context = drop_dict(context, "date", inplace=False)
        elif len(date_range) > 1:
            period = [dict(startDate=start, endDate=(end-dt.timedelta(days=1)))
                        for start, end in zip(date_range, date_range[1:]+[endDate+dt.timedelta(days=1)])]
        else: period = [dict(startDate=startDate, endDate=endDate)]
        return period, context

    def product_context(self, *context: Sequence[Context], **kawrgs) -> List[Context]:
        context_array = map((lambda x: x if x else [{}]), context)
        return list(map(chain_dict, product(*context_array)))

    ###################################################################
    ########################## Fetch Request ##########################
    ###################################################################

    @abstractmethod
    @log_errors
    @requests_limit
    def fetch(self, *args, **kwargs) -> Data:
        ...

    def request_content(self, method: str, url: str, session: Optional[requests.Session]=None,
                        params=None, data=None, json=None, headers=None, cookies=None,
                        allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> bytes:
        session = session if session else requests
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(log_response(response, url=url, **logQuery))
            return response.content

    def request_text(self, method: str, url: str, session: Optional[requests.Session]=None,
                    params=None, data=None, json=None, headers=None, cookies=None,
                    allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> str:
        session = session if session else requests
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(log_response(response, url=url, **logQuery))
            return response.text

    def request_json(self, method: str, url: str, session: Optional[requests.Session]=None,
                    params=None, data=None, json=None, headers=None, cookies=None,
                    allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> JsonData:
        session = session if session else requests
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(log_response(response, url=url, **logQuery))
            return response.json()

    def request_headers(self, method: str, url: str, session: Optional[requests.Session]=None,
                        params=None, data=None, json=None, headers=None, cookies=None,
                        allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> Dict:
        session = session if session else requests
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(log_response(response, url=url, **logQuery))
            return response.headers

    def extra_save(self, data: Data, prefix=str(), extraSave=False, rename: RenameDict=dict(), **kwargs):
        if not extraSave: return
        file = prefix+'_' if prefix else str()+now("%Y%m%d%H%M%S")+".xlsx"
        convert_data(data, returnType="dataframe").rename(columns=rename).to_excel(file, index=False)

    ###################################################################
    ############################ Google API ###########################
    ###################################################################

    def gcloud_authorized(func):
        @functools.wraps(func)
        async def wrapper(self: Spider, *args, redirectUrl=str(), authorization=str(), account: Account=dict(), **kwargs):
            if not authorization:
                authorization = fetch_gcloud_authorization(redirectUrl, account, **kwargs)
            return await func(
                self, *args, redirectUrl=redirectUrl, authorization=authorization, account=account, **kwargs)
        return wrapper

    def set_query(self, queryKey: Keyword=str(), querySheet: Keyword=str(), queryFields: IndexLabel=list(),
                    queryString: NumericiseIgnore=list(), queryArray: IndexLabel=list(),
                    rename: Dict[str,str]=dict(), account: Account=dict(), **context):
        if not (queryKey and querySheet and queryFields): return
        elif is_same_length(queryKey, querySheet, queryFields):
            queryString = fill_array(queryString, count=len(queryKey), value=list())
            queryArray = fill_array(queryArray, count=len(queryKey), value=list())
            for key, sheet, fields, str_cols in zip(queryKey, querySheet, queryFields, queryString, queryArray):
                self.set_gs_query(key, sheet, fields, str_cols, rename, account=account, **context)
        elif isinstance(queryKey, str) and isinstance(querySheet, str):
            self.set_gs_query(queryKey, querySheet, queryFields, queryString, queryArray, rename, account=account, **context)

    def upload_data(self, data: TabularData, gsKey: Keyword=str(), gsSheet: Keyword=str(), gsMode: Keyword="append",
                    gsBaseSheet: Keyword=str(), gsRange: Keyword=str(), rename: RenameDict=dict(),
                    gbqPid: Keyword=str(), gbqTable: Keyword=str(), gbqMode: Keyword="append",
                    gbqSchema: Optional[SchemaSequence]=None, gbqProgress=True,
                    gbqPartition: Keyword=str(), gbqPartitionBy: Keyword="auto",
                    gbqReauth=False, account: Account=dict(), credentials: IDTokenCredentials=None, **context):
        if data_empty(data) or not ((gsKey and gsSheet) or (gbqPid and gbqTable)): return
        data = convert_data(data, returnType="dataframe")
        if gsKey and gsSheet and (isinstance(gsKey, str) and isinstance(gsSheet, str)) or is_same_length(gsKey, gsSheet):
            gs_args = tuple(map(cast_tuple, [gsKey, gsSheet, gsMode, gsBaseSheet, gsRange]))
            for key, sheet, mode, base_sheet, cell in zip(*align_array(gs_args, how="first", default=str())):
                self.upload_gspread(key, sheet, data.copy(), (mode if mode else "append"), base_sheet, cell,
                                    rename=rename, account=account, **context)
        if gbqPid and gbqTable and (isinstance(gbqPid, str) and isinstance(gbqTable, str)) or is_same_length(gbqPid, gbqTable):
            gbq_args = tuple(map(cast_tuple, [gbqPid, gbqTable, gbqMode, gbqSchema, gbqPartition, gbqPartitionBy]))
            for pid, table, mode, schema, partition, partition_by in zip(*align_array(gbq_args, how="first", default=str())):
                self.upload_gbq(table, pid, data.copy(), (mode if mode else "append"), schema, gbqProgress,
                                partition, partition_by, reauth=gbqReauth, account=account, credentials=credentials, **context)

    ###################################################################
    ########################## Google Spread ##########################
    ###################################################################

    def set_gs_query(self, key: str, sheet: str, fields: IndexLabel, str_cols: NumericiseIgnore=list(),
                    arr_cols: IndexLabel=list(), rename: RenameDict=dict(), account: Account=dict(), **kwargs):
        data = read_gspread(key, sheet, account, numericise_ignore=str_cols, rename=rename)
        self.logger.info(log_table(data, logJson=self.logJson))
        self.update(**{field:data[field].tolist() for field in cast_tuple(fields) if field in data})
        for field in cast_tuple(fields):
            values = data[field].tolist() if field in data else list()
            if (len(values) == 1) and field not in arr_cols: values = values[0]
            self.update(field=values)

    @log_errors
    def upload_gspread(self, key: str, sheet: str, data: pd.DataFrame, mode="append", base_sheet=str(),
                        cell=str(), rename: RenameDict=dict(), account: Account=dict(), clear=False, **kwargs):
        data = self.map_gs_data(data, **kwargs)
        if base_sheet:
            data = self.map_gs_base(data, self.read_gs_base(key, base_sheet, rename=rename, account=account), **kwargs)
        self.logger.info(log_table(data, key=key, sheet=sheet, logJson=self.logJson))
        cell, clear = ("A2" if mode == "replace" else (cell if cell else str())), (True if mode == "replace" else clear)
        update_gspread(key, sheet, data, account=account, cell=cell, clear=clear)

    def map_gs_data(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        return data

    def read_gs_base(self, key: str, sheet: str, str_cols: NumericiseIgnore=list(),
                    rename: RenameDict=dict(), account: Account=dict(), **kwargs) -> pd.DataFrame:
        data = read_gspread(key, sheet, account=account, numericise_ignore=str_cols, rename=rename)
        self.logger.info(log_table(data, key=key, sheet=sheet, logJson=self.logJson))
        return data

    def map_gs_base(self, data: pd.DataFrame, base: pd.DataFrame, **kwargs) -> pd.DataFrame:
        return cloc(data, base.columns, default="pass")

    ###################################################################
    ######################### Google Bigquery #########################
    ###################################################################

    @log_errors
    def upload_gbq(self, table: str, project_id: str, data: pd.DataFrame, mode="append",
                    schema: Optional[BigQuerySchema]=None, progress=True, partition=str(), partition_by="auto",
                    reauth=False, account: Account=dict(), credentials: Optional[IDTokenCredentials]=None, **kwargs):
        schema = schema if schema and is_records(schema) else self.get_gbq_schema(mode=mode, schema=schema, **kwargs)
        data = self.map_gbq_data(data, schema=schema, **kwargs)
        context = dict(project_id=project_id, reauth=reauth, account=account, credentials=credentials)
        if mode == "upsert":
            data = self.map_gbq_base(data, self.read_gbq_base(query=table, **context), schema=schema, **kwargs)
        self.logger.info(log_table(data, table=table, pid=project_id, mode=mode, schema=schema, logJson=self.logJson))
        to_gbq(table, project_id, data, if_exists=("replace" if mode == "upsert" else mode), schema=schema, progress=progress,
                partition=partition, partition_by=partition_by, **kwargs)

    def get_gbq_schema(self, mode="append", schema=str(), **kwargs) -> BigQuerySchema:
        ...

    def map_gbq_data(self, data: pd.DataFrame, schema: Optional[BigQuerySchema]=None, **kwargs) -> pd.DataFrame:
        columns = (field.get("name") for field in schema) if schema else tuple()
        return cloc(data, columns=columns, default="pass")

    def read_gbq_base(self, query: str, project_id: str, **kwargs) -> pd.DataFrame:
        return read_gbq(query, project_id, **kwargs)

    def map_gbq_base(self, data: pd.DataFrame, base: pd.DataFrame, key=str(),
                    schema: Optional[BigQuerySchema]=None, **kwargs) -> pd.DataFrame:
        key = validate_upsert_key(data, base, key, schema)
        data = apply_df(data, apply=(lambda x: None if x == str() else x), all_cols=True)
        data = data.set_index(key).combine_first(base.set_index(key)).reset_index()
        return data[data[key].notna()].drop_duplicates(key)


###################################################################
########################## Async Spiders ##########################
###################################################################

class AsyncSpider(Spider):
    __metaclass__ = ABCMeta
    asyncio = True
    operation = "asyncSpider"
    where = "urls"
    which = "data"
    contextQuery = list()
    redirectQuery = list()
    returnType = "records"
    errors = defaultdict(list)
    maxLimit = MAX_ASYNC_TASK_LIMIT
    redirectLimit = MAX_REDIRECT_LIMIT
    message = str()
    rename = dict()

    ###################################################################
    ########################## Async Managers #########################
    ###################################################################

    def asyncio_task(func):
        @functools.wraps(func)
        async def wrapper(self: AsyncSpider, *args, self_var=True, **kwargs):
            if self_var: kwargs = dict(self.__dict__, **kwargs)
            semaphore = self.asyncio_semaphore(**kwargs)
            data = await func(self, *args, semaphore=semaphore, **kwargs)
            self.upload_data(data, **GCP_CONTEXT(**kwargs))
            return data
        return wrapper

    def asyncio_session(func):
        @functools.wraps(func)
        async def wrapper(self: AsyncSpider, *args, self_var=True, **kwargs):
            if self_var: kwargs = dict(self.__dict__, **kwargs)
            semaphore = self.asyncio_semaphore(**kwargs)
            async with aiohttp.ClientSession() as session:
                data = await func(self, *args, session=session, semaphore=semaphore, **kwargs)
            await asyncio.sleep(.25)
            self.upload_data(data, **GCP_CONTEXT(**kwargs))
            return data
        return wrapper

    def asyncio_semaphore(self, numTasks: Optional[int]=None, apiRedirect=False, **kwargs) -> asyncio.Semaphore:
        numTasks = min(cast_int(numTasks, default=self.maxLimit), (self.redirectLimit if apiRedirect else self.maxLimit))
        return asyncio.Semaphore(numTasks)

    def asyncio_errors(func):
        @functools.wraps(func)
        async def wrapper(self: AsyncSpider, *args, **kwargs):
            try: return await func(self, *args, **kwargs)
            except KeyboardInterrupt as interrupt:
                raise interrupt
            except Exception as exception:
                if self.logErrors: self.log_arguments(args, kwargs)
                if self.debug: raise exception
                name = f"{func.__name__}({self.__class__.__name__})"
                self.logger.error(log_exception(name, *args, json=self.logJson, **kwargs))
                return init_origin(func)
        return wrapper

    def asyncio_redirect(func):
        @functools.wraps(func)
        async def wrapper(self: AsyncSpider, *args, apiRedirect=False, **kwargs):
            if apiRedirect: return await self.redirect(self, *args, **kwargs)
            else: return await func(self, *args, **kwargs)
        return wrapper

    def asyncio_limit(func):
        @functools.wraps(func)
        async def wrapper(self: AsyncSpider, *args, semaphore: Optional[asyncio.Semaphore]=None, **kwargs):
            if self.delay:
                await asyncio.sleep(self.delay)
            if semaphore:
                async with semaphore:
                    return await func(self, *args, **kwargs)
            return await func(self, *args, **kwargs)
        return wrapper

    ###################################################################
    ########################## Async Requests #########################
    ###################################################################

    @abstractmethod
    @asyncio_session
    async def crawl(self, **context) -> Data:
        return await self.gather(**self.map_context(**context))

    @asyncio_redirect
    async def gather(self, *args, message=str(), progress=None, filter: IndexLabel=list(),
                    returnType: Optional[TypeHint]=None, **context) -> Data:
        message = message if message else GATHER_MSG(self.which, self.where)
        hide_bar = not (progress if isinstance(progress, bool) else self.progress)
        query, context = self.split_context(*args, **context)
        data = await tqdm.gather(*[
                self.fetch(**__q, filter=filter, **context) for __q in query], desc=message, disable=hide_bar)
        return self.map_reduce(data, filter=filter, returnType=returnType, **context)

    @abstractmethod
    @asyncio_errors
    @asyncio_limit
    async def fetch(self, *args, **kwargs) -> Data:
        ...

    async def request_content(self, method: str, url: str, session: Optional[aiohttp.ClientSession]=None,
                        params=None, data=None, json=None, headers=None, cookies=None,
                        allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> bytes:
        session = session if session else aiohttp
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        async with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(await log_client(response, url=url, **logQuery))
            return await response.read()

    async def request_text(self, method: str, url: str, session: Optional[aiohttp.ClientSession]=None,
                    params=None, data=None, json=None, headers=None, cookies=None,
                    allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> str:
        session = session if session else aiohttp
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        async with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(await log_client(response, url=url, **logQuery))
            return await response.text()

    async def request_json(self, method: str, url: str, session: Optional[aiohttp.ClientSession]=None,
                    params=None, data=None, json=None, headers=None, cookies=None,
                    allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> JsonData:
        session = session if session else aiohttp
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        async with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(await log_client(response, url=url, **logQuery))
            return await response.json()

    async def request_headers(self, method: str, url: str, session: Optional[aiohttp.ClientSession]=None,
                        params=None, data=None, json=None, headers=None, cookies=None,
                        allow_redirects=True, logQuery: Optional[Dict]=dict(), **kwargs) -> Dict:
        session = session if session else aiohttp
        messages = dict(params=params, data=data, json=json, headers=headers, cookies=cookies)
        self.logger.debug(log_messages(**messages, logJson=self.logJson))
        async with session.request(method, url, **messages, allow_redirects=allow_redirects) as response:
            self.logger.info(await log_client(response, url=url, **logQuery))
            return response.headers

    ###################################################################
    ########################## Async Redirect #########################
    ###################################################################

    @Spider.gcloud_authorized
    async def redirect(self, *args, contextUnit: Unit=1, redirectUnit: Unit=1, message=str(), progress=None,
                        filter: IndexLabel=list(), returnType: Optional[TypeHint]=None, **context) -> Data:
        message = message if message else GATHER_MSG(self.which, self.where)
        hide_bar = not (progress if isinstance(progress, bool) else self.progress)
        query, context = self.split_context(*args, contextUnit=redirectUnit, **context)
        data = await tqdm.gather(*[
                self.fetch_redirect(**__q, filter=filter, **context) for __q in query], desc=message, disable=hide_bar)
        return self.map_reduce(data, filter=filter, returnType=returnType, **context)

    @asyncio_errors
    @asyncio_limit
    @Spider.gcloud_authorized
    async def fetch_redirect(self, redirectUrl: str, authorization: str, session: Optional[aiohttp.ClientSession]=None,
                            account: Account=dict(), **context) -> Records:
        data = self.filter_redirect_data(redirectUrl, authorization, account, **context)
        response = self.request_text("POST", redirectUrl, session, json=data, headers=dict(Authorization=authorization))
        return self.parse_redirect(json.loads(response), **context)

    def filter_redirect_data(self, redirectUrl: str, authorization: str, account: Account=dict(),
                            redirectFilter: Optional[ContextMapper]=None, **context) -> Context:
        redirectFilter = redirectFilter if redirectFilter else (lambda **context: context)
        return dict(
            redirectUrl=redirectUrl,
            authorization=authorization,
            account=account,
            **REDIRECT_CONTEXT(**redirectFilter(**context)))

    def parse_redirect(self, data: RedirectData, **kwargs) -> Records:
        if self.redirectErrors:
            data = self.log_redirect_errors(data)
        data = self.map_redirect(data, **kwargs)
        self.logger.debug(log_data(data))
        return data

    def log_redirect_errors(self, data: RedirectData, **kwargs) -> Records:
        if not isinstance(data, Dict): raise ValueError(INVALID_REDIRECT_LOG_MSG)
        errors = data.get("errors", default=dict())
        for key, values in (errors if isinstance(errors, Dict) else dict()).items():
            self.errors[key] += values
        return data.get("data", default=list())

    def map_redirect(self, data: Records, **kwargs) -> Records:
        if not isinstance(data, List): return list()
        cast_datetime_or_keep = lambda x: cast_datetime_format(x, default=x)
        return [apply_df(__m, apply=cast_datetime_or_keep, all_cols=True) for __m in data]

###################################################################
######################## Encrypted Spiders ########################
###################################################################

class LoginSpider(requests.Session, Spider):
    __metaclass__ = ABCMeta
    asyncio = False
    operation = "login"
    where = str()

    @abstractmethod
    def __init__(self, logName=str(), logLevel: Union[str,int]="WARN", logFile: Optional[str]=str(), **kwargs):
        super().__init__()
        self.operation = self.operation
        self.logName = logName if logName else self.operation
        self.logLevel = int(logLevel) if str(logLevel).isdigit() else logging.getLevelName(str(logLevel).upper())
        self.logFile = logFile
        self.logJson = bool(logFile)
        self.logger = CustomLogger(name=self.logName, level=self.logLevel, file=self.logFile)

    @abstractmethod
    def login(self):
        ...

    def login_required(func):
        @functools.wraps(func)
        def wrapper(self: LoginSpider, *args, **kwargs):
            context = self.get_user_info(**kwargs)
            return func(self, *args, **kwargs, **context)
        return wrapper

    def get_user_info(self, cookies=str(), **kwargs) -> Context:
        cookies = cookies if cookies else self.get_cookies()
        if not cookies:
            self.login()
            cookies = self.get_cookies()
        if not cookies: raise KeyboardInterrupt(LOGIN_REQUIRED_MSG(self.where))
        else: return dict(cookies=cookies)

    def get_cookies(self, **kwargs) -> str:
        return parse_cookies(self.cookies)


class EncryptedSpider(Spider):
    __metaclass__ = ABCMeta
    asyncio = False
    operation = "encryptedSpider"
    where = "urls"
    which = "data"
    contextQuery = list()
    redirectQuery = list()
    returnType = "records"
    errors = defaultdict(list)
    message = str()
    rename = dict()
    authClass = LoginSpider
    idKey = "userid"
    pwKey = "passwd"
    extraKeys = list()

    def set_context(self, cookies=str(), encryptedKey=str(), filterContext: Optional[ContextMapper]=None, **context):
        self.cookies = cookies
        self.update(filterContext(**context) if isinstance(filterContext, Callable) else context)
        self.set_secrets(encryptedKey)

    def set_secrets(self, encryptedKey: str):
        self.encryptedKey = encryptedKey
        decryptedKey = {key:self.get(key, default=str()) for key in [self.idKey, self.pwKey]+self.extraKeys}
        self.decryptedKey = json.loads(decrypt(encryptedKey,1)) if encryptedKey else decryptedKey
        if not encryptedKey and any(map(len,decryptedKey.values())):
            self.update(encryptedKey=encrypt(json.dumps(decryptedKey, ensure_ascii=False, default=1),1))
        self.logger.info(log_encrypt(**self.decryptedKey))

    def login(self, update=True) -> str:
        auth = self.authClass(logFile=self.logFile, logLevel=self.logLevel, **self.decryptedKey)
        auth.login()
        cookies = auth.get_cookies()
        auth.close()
        return self.update(cookies=cookies) if update else cookies

    def login_required(func):
        @functools.wraps(func)
        def wrapper(self: EncryptedSpider, *args, **kwargs):
            context = self.get_user_info(**kwargs)
            return func(self, *args, **kwargs, **context)
        return wrapper

    def get_user_info(self, cookies=str(), **kwargs) -> Context:
        cookies = cookies if cookies else self.cookies
        if not cookies:
            self.login(update=True)
            cookies = self.cookies
        if not cookies: raise KeyboardInterrupt(LOGIN_REQUIRED_MSG(self.where))
        else: return dict(cookies=cookies)

    def validate_response(self, response: requests.Response, invalid_status: Sequence[int]=(401,), **kwargs):
        if response.status_code in invalid_status:
            raise KeyboardInterrupt(LOGIN_INVALID_MSG(self.where))


class EncryptedAsyncSpider(AsyncSpider, EncryptedSpider):
    __metaclass__ = ABCMeta
    asyncio = True
    operation = "encryptedAsyncSpider"
    where = "urls"
    which = "data"
    contextQuery = list()
    redirectQuery = list()
    returnType = "records"
    errors = defaultdict(list)
    message = str()
    rename = dict()
    authClass = LoginSpider
    idKey = "userid"
    pwKey = "passwd"
    extraKeys = list()

    def login_required(func):
        @functools.wraps(func)
        async def wrapper(self: EncryptedAsyncSpider, *args, cookies=str(), **kwargs):
            context = self.get_user_info(**kwargs)
            return await func(self, *args, **kwargs, **context)
        return wrapper

    def validate_response(self, response: aiohttp.ClientResponse, invalid_status: Sequence[int]=(401,), **kwargs):
        if response.status in invalid_status:
            raise KeyboardInterrupt(LOGIN_INVALID_MSG(self.where))


###################################################################
############################# Parsers #############################
###################################################################

class Parser(CustomDict):
    __metaclass__ = ABCMeta
    operation = "parser"

    def __init__(self, logName=str(), logLevel="WARN", logFile=str(), **kwargs):
        super().__init__()
        self.logName = logName if logName else self.operation
        self.logLevel = int(logLevel) if str(logLevel).isdigit() else logging.getLevelName(str(logLevel).upper())
        self.logFile = logFile
        self.logJson = bool(logFile)
        self.logger = CustomLogger(name=self.logName, level=self.logLevel, file=self.logFile)

    @abstractmethod
    def parse(self, response: Any, **kwargs) -> Data:
        ...


###################################################################
############################ Pipelines ############################
###################################################################

class Pipeline(Spider):
    __metaclass__ = ABCMeta
    asyncio = False
    operation = "pipeline"
    dependencies = (Spider, Spider)
    contextQuery = list()
    redirectQuery = list()
    returnType = "dataframe"
    errors = defaultdict(list)
    rename = dict()

    @abstractmethod
    @Spider.requests_task
    def crawl(self, **context) -> Data:
        return self.gather(**context)

    @abstractmethod
    def gather(self, **context) -> Data:
        args = list()
        for cralwer in self.dependencies:
            operation = self.get_operation(cralwer, **context)
            data = self.crawl_proxy(cralwer, prefix=operation, **context)
            args.append(data)
        return self.map_reduce(*args, **context)

    def get_operation(self, crawler: Spider, rename: RenameDict=dict(), **kwargs) -> str:
        operation = crawler.__dict__.get("operation")
        if not operation: raise ValueError(DEPENDENCY_HAS_NO_NAME_MSG)
        return rename.get(operation, default=operation)

    def crawl_proxy(self, crawler: Spider, prefix=str(), extraSave=False,
                    appendix: Optional[pd.DataFrame]=None, drop="right", how="left", on=str(), **context) -> Data:
        query = crawler.contextArgs+crawler.contextQuery
        if query and all(map(pd.notna, kloc(context, query, value_only=True))): return pd.DataFrame()
        crawler = crawler(**context)
        data = pd.DataFrame(crawler.crawl(**PROXY_CONTEXT(**crawler.__dict__)))
        self.extra_save(data, prefix=prefix, extraSave=extraSave, **context)
        return merge_drop(data, appendix, drop=drop, how=how, on=on)

    @abstractmethod
    def map_reduce(self, *data: Data, filter=list(), returnType: Optional[TypeHint]=None, **kwargs) -> Data:
        return filter_data(chain_exists(data), filter=filter, returnType=returnType)


class AsyncPipeline(AsyncSpider, Pipeline):
    __metaclass__ = ABCMeta
    operation = "asyncPipeline"
    dependencies = (AsyncSpider, AsyncSpider)
    contextQuery = list()
    redirectQuery = list()
    returnType = "dataframe"
    errors = defaultdict(list)
    rename = dict()

    @abstractmethod
    @AsyncSpider.asyncio_task
    async def crawl(self, **context) -> Data:
        return await self.gather(**context)

    @abstractmethod
    async def gather(self, **context) -> Data:
        args = list()
        for cralwer in self.dependencies:
            operation = self.get_operation(cralwer, **context)
            if cralwer.asyncio: data = await self.async_proxy(cralwer, prefix=operation, **context)
            else: data = self.crawl_proxy(cralwer, prefix=operation, **context)
            args.append(data)
        return self.map_reduce(*args, **context)

    async def async_proxy(self, crawler: Spider, prefix=str(), extraSave=False,
                            appendix: Optional[pd.DataFrame]=None,
                            drop="right", how="left", on=str(), **context) -> Data:
        query = crawler.contextArgs+crawler.contextQuery
        if query and all(map(pd.notna, kloc(context, query, value_only=True))): return pd.DataFrame()
        crawler = crawler(**context)
        data = pd.DataFrame(await crawler.crawl(**PROXY_CONTEXT(**crawler.__dict__)))
        self.extra_save(data, prefix=prefix, extraSave=extraSave, **context)
        return merge_drop(data, appendix, drop=drop, how=how, on=on)


###################################################################
########################### Log Managers ##########################
###################################################################

def log_encrypt(show=3, **kwargs) -> LogMessage:
    encrypt = lambda string, show=3: str(string[:show]).ljust(len(string),'*')
    return dict(**{key:encrypt(value, show=show) for key, value in kwargs.items()})


def log_messages(params: Optional[Dict]=None, data: Optional[Dict]=None, json: Optional[Dict]=None,
                headers: Optional[Dict]=None, cookies=None, logJson=False, **kwargs) -> LogMessage:
    dumps = lambda struct: dumps_map(struct) if logJson else str(struct)
    params = {"params":dumps(params)} if params else dict()
    data = {"data":dumps(data if data else json)} if data or json else dict()
    headers = {"headers":dumps(headers)} if headers else dict()
    cookies = {"cookies":dumps(cookies)} if cookies else dict()
    kwargs = {key:dumps(values) for key,values in kwargs.items()}
    return dict(**kwargs, **data, **params, **headers, **cookies)


def log_response(response: requests.Response, url: str, **kwargs) -> LogMessage:
    try: length = len(response.text)
    except: length = None
    return dict(**kwargs, **{"url":unquote(url), "status":response.status_code, "contents-length":length})


async def log_client(response: aiohttp.ClientResponse, url: str, **kwargs) -> LogMessage:
    try: length = len(await response.text())
    except: length = None
    return dict(**kwargs, **{"url":unquote(url), "status":response.status, "contents-length":length})


def log_data(results: List, **kwargs) -> LogMessage:
    try: length = len(results)
    except: length = None
    return dict(**kwargs, **{"data-length":length})


def log_exception(func: str, *args, logJson=False, **kwargs) -> LogMessage:
    dumps = lambda struct: (dumps_map(struct) if logJson else str(struct))[:1000]
    error = ''.join(traceback.format_exception(*sys.exc_info()))
    error = unraw(error) if logJson else error
    return dict(func=func, args=dumps(args), kwargs=dumps(kwargs), error=error)


def log_table(data: pd.DataFrame, schema: Optional[BigQuerySchema]=None, logJson=False, **kwargs) -> LogMessage:
    dumps = lambda struct: dumps_map(struct) if logJson else str(struct)
    schema = {"schema":dumps(schema)} if schema else dict()
    try: shape = data.shape
    except: shape = (0,0)
    return dict(**kwargs, **{"table-shape":str(shape)}, **schema)



from .cast import cast_list, cast_tuple, cast_str
from .types import _KT, _VT, _bool, _type, Comparable, Index, IndexLabel, Keyword, TypeHint
from .types import Context, NestedSequence, IndexedSequence, Records, TabularData, MappingData, Data
from .types import ApplyFunction, MatchFunction, BetweenRange
from .types import is_int_array, is_array, is_2darray, is_records, is_dfarray
from .types import is_list_type,is_dict_type, is_records_type, is_dataframe_type

from typing import Any, Dict, List, Set
from typing import Callable, Iterable, Optional, Sequence, Tuple, Union
from itertools import chain
from functools import cmp_to_key, reduce
import pandas as pd
import re


INCLUDE = 0
EXCLUDE = 1

exists = lambda x: bool(x) and (pd.notna(x) if not isinstance(x, Sequence) else True)
is_empty = lambda x: not x or (pd.isna(x) if not isinstance(x, Sequence) else False)
str_na = lambda x: str(x) if not isinstance(x, Sequence) and pd.notna(x) else str()

df_exists = lambda df, null=True: (not (df if null else df.dropna(axis=0, how="all")).empty) if isinstance(df, pd.DataFrame) else False
df_empty = lambda df, null=True: (df if null else df.dropna(axis=0, how="all")).empty if isinstance(df, pd.DataFrame) else False

data_exists = lambda data: not data.empty if isinstance(data, pd.DataFrame) else bool(data)
data_empty = lambda data: data.empty if isinstance(data, pd.DataFrame) else not bool(data)

__or = lambda *args: reduce(lambda x,y: x|y, args)
__and = lambda *args: reduce(lambda x,y: x&y, args)
union = lambda *arrays: reduce(lambda x,y: x+y, arrays)
inter = lambda *arrays: reduce(lambda x,y: [e for e in x if e in y], arrays)


def __apply(__object, __applyFunc: ApplyFunction, default=None, **kwargs) -> Any:
    try: return __applyFunc(__object, **kwargs)
    except: return default


def allin(__object: Iterable) -> bool:
    return bool(tuple(__object)) and all(__object)


INCLUSIVE = ("both", "neither", "left", "right")

def between(__object: Comparable, left=None, right=None, inclusive="both", **kwargs) -> bool:
    match_left = ((left == None) or (__object >= left if inclusive in ["both","left"] else __object > left))
    match_right = ((right == None) or (__object <= right if inclusive in ["both","right"] else __object < right))
    return match_left & match_right


def map_context(__keys: _KT, __values: _VT, **context) -> Context:
    if context: return context
    if not (__keys and __values): return dict()
    __keys = cast_tuple(__keys)
    if not is_array(__values): __values = (__values,)*len(__keys)
    elif len(__keys) != len(__values): return dict()
    return {__key:__value for __key, __value in zip(__keys, __values)}


###################################################################
############################## String #############################
###################################################################

def re_get(pattern: str, string: str, default=str(), groups=False, **kwargs) -> str:
    if not re.search(pattern, string): return default
    catch = re.search(pattern, string).groups()
    return catch[0] if catch and not groups else catch


def replace_map(string: str, __m: dict, **kwargs) -> str:
    for __old, __new in __m.items():
        string = string.replace(__old, __new)
    return string


###################################################################
############################## Array ##############################
###################################################################

abs_idx = lambda idx: abs(idx+1) if idx < 0 else idx


def iloc(__s: IndexedSequence, index: Index, default=None, **kwargs) -> _VT:
    length = len(__s)
    if isinstance(index, int):
        return __s[index] if abs_idx(index) < length else (default if default != "pass" else None)
    elif not is_int_array(index, empty=False): return __s
    elif default == "pass": return [__s[i] for i in index if abs_idx(i) < length]
    else: return [__s[i] if abs_idx(i) < length else default for i in index]


def flatten(*args, iter_type: Tuple[_type]=(List,Set,Tuple), **kwargs) -> List:
    return [__e for __object in args for __e in (
            __object if isinstance(__object, iter_type) else cast_tuple(__object))]


def unique(*elements, empty=False, null_only=False, **kwargs) -> List:
    array = list()
    is_valid = pd.notna if null_only else exists
    for element in elements:
        if not (empty or is_valid(element)): continue
        if element not in array: array.append(element)
    return array


def apply_array(__s: IndexedSequence, __indices: Optional[Index]=list(), __applyFunc: Optional[ApplyFunction]=list(),
                all_indices=False, apply: Optional[ApplyFunction]=None, **context) -> IndexedSequence:
    __s = __s.copy()
    if all_indices:
        if not isinstance(apply, Callable): return __s
        else: return [apply(__e) for __e in __s]
    for __index, applyFunc in map_context(__indices, __applyFunc, **context).items():
        if isinstance(applyFunc, Callable): continue
        elif not isinstance(__index, int): __s = applyFunc(__s)
        elif abs_idx(__index) not in __s: continue
        else: __s[__index] = applyFunc(__s[__index])
    return __s


def match_array(__s: IndexedSequence, __indices: Optional[Index]=list(), __matchFunc: Optional[ApplyFunction]=list(),
                all_indices=False, **context) -> bool:
    if all_indices:
        if not isinstance(__matchFunc, Callable): return False
        else: return __and(map(__matchFunc, __s))
    match = True
    for __index, matchFunc in map_context(__indices, __matchFunc, **context).items():
        if not isinstance(matchFunc, Callable): continue
        elif not isinstance(__index, int): match &= matchFunc(__s)
        elif abs_idx(__index) not in __s: continue
        else: match &= matchFunc(__s[__index])
    return match


def chain_array(__s: NestedSequence, empty=True, **kwargs) -> List:
    if not __s: return list()
    elif not empty: __s = [element for element in __s if exists(element)]
    return list(chain.from_iterable(__s))


def get_index(__s: IndexedSequence, values: _VT, default=None, multiple=True, **kwargs) -> _VT:
    if not (is_array(values) and multiple):
        return __s.index(values) if values in __s else (default if default != "pass" else None)
    elif not values: return __s
    elif default == "pass": return [__s.index(value) for value in values if value in __s]
    else: return [__s.index(value) if value in __s else default for value in values]


def get_scala(__object, index: Optional[int]=None, default=None, **kwargs) -> _VT:
    if not (__object and is_array(__object)): return __object
    elif isinstance(__object, Set): return __object.copy().pop()
    else: return iloc(__object, (index if isinstance(index, int) else 0), default=default)


def values_to_back(__s: Sequence, values: _VT, **kwargs) -> List:
    values = cast_tuple(values)
    return sorted(__s, key=cmp_to_key(lambda x, y: -1 if y in values else 0))


def fill_array(__s: Sequence, count: int, value=None, **kwargs) -> List:
    return [__s[i] if i < len(__s) else value for i in range(count)]


def filter_array(__s: Sequence, match: MatchFunction, apply: Optional[ApplyFunction]=None, **kwargs) -> List:
    if apply: return [apply(element) for element in __s if match(element)]
    else: [element for element in __s if match(element)]


def is_same_length(*args: Sequence, how="all", **kwargs) -> bool:
    __l = set()
    for idx, __object in enumerate(args):
        if isinstance(__object, Sequence): __l.add(len(__object))
        elif idx != 0 and how == "first": pass
        else: return False
    return len(__l) <= 1


def align_index(*__s: Sequence, how="all", unique=False, null_only=False, **kwargs) -> List[int]:
    __iter, __set, count = __s, [set() for _ in __s], 0
    is_valid = pd.notna if null_only else exists
    if len(__iter) > 0 and how == "all":
        count = min([len(array) for array in __iter])
    elif len(__iter) > 0 and how == "first":
        count = len(__iter[0])
        __iter, __set = [__s[0]], [set()]
    if unique:
        index = [set() for _ in __iter]
        for cur, array in enumerate(__iter):
            for idx, element in enumerate(array):
                if is_empty(element) or element in __set[cur]: continue
                index[cur].add(idx)
                __set[cur].add(element)
        return sorted(__and(*index))
    else: return [idx for idx in range(count) if all([is_valid(array[idx]) for array in __iter])]


def align_array(*__s: Sequence, how="all", default: Optional[Union[Any,Sequence]]=None, empty=True,
                unique=False, null_only=False, **kwargs) -> Tuple[List]:
    is_valid = pd.notna if null_only else exists
    get_default = (lambda idx: iloc(default, idx)) if is_array(default) and default else (lambda idx: default)
    if len(__s) == 1:
        return __s[0] if empty else filter_array(__s[0], is_valid)
    elif len(__s) > 1 and how == "all" and empty:
        return tuple(fill_array(array, max([len(array) for array in __s]), get_default(i)) for i, array in enumerate(__s))
    elif len(__s) > 1 and how == "first" and empty:
        return tuple([__s[0]]+[fill_array(array, len(__s[0]), get_default(i)) for i, array in enumerate(__s[1:])])
    elif len(__s) > 1 and how == "dropna":
        if len(set(map(len,__s))) != 1: return __s
        matches = [idx for idx in range(len(__s[0])) if any(map(lambda x: is_valid(x[idx]), __s))]
        return tuple([array[:max(matches)+1] if matches else list() for array in __s])
    valid = align_index(*__s, how=how, unique=unique, null_only=null_only,)
    return tuple([array[idx] if idx < len(array) else get_default(i) for idx in valid] for i, array in enumerate(__s))


###################################################################
############################### Map ###############################
###################################################################

def kloc(__m: Dict, __keys: _KT, default=None, value_only=False, match_type=False, **kwargs) -> Union[_VT,Dict]:
    cast = default.__class__ if match_type and default != None else (lambda x: x)
    if not is_array(__keys): return __m.get(__keys, cast(default if default != "pass" else None))
    elif __keys and value_only:
        if default == "pass": return [cast(__m[key]) for key in __keys if key in __m]
        else: return [cast(__m.get(key, default)) for key in __keys]
    elif __keys and not value_only:
        if default == "pass": return {key:cast(__m[key]) for key in __keys if key in __m}
        else: return {key:cast(__m.get(key, default)) for key in __keys}
    else: return __m


def apply_dict(__m: Dict, __keys: Optional[_KT]=list(), __applyFunc: Optional[ApplyFunction]=list(),
                all_keys=False, apply: Optional[ApplyFunction]=None, **context) -> Dict:
    __m = __m.copy()
    if all_keys:
        if not isinstance(apply, Callable): return __m
        return {__key:apply(__values) for __key, __values in __m.items()}
    for __key, applyFunc in map_context(__keys, __applyFunc, **context).items():
        if not isinstance(applyFunc, Callable): continue
        elif not __key: __m = applyFunc(__m)
        elif __key not in __m: continue
        else: __m[__key] = applyFunc(__m[__key])
    return __m


def match_dict(__m: Dict, __keys: Optional[_KT]=list(), __matchFunc: Optional[MatchFunction]=list(),
                all_keys=False, **context) -> bool:
    if all_keys:
        if not isinstance(__matchFunc, Callable): return False
        else: return __and(map(__matchFunc, __m.values()))
    match = True
    for __key, matchFunc in map_context(__keys, __matchFunc, **context).items():
        if not isinstance(matchFunc, Callable): continue
        elif not __key: match &= matchFunc(__m)
        elif __key not in __m: continue
        else: match &= matchFunc(__m[__key])
    return match


def diff_dict(__m: Dict, **kwargs) -> Dict:
    return {key:value for key,value in __m.items() if key not in kwargs}


def chain_dict(__object: Sequence[Dict], keep="first", empty=True, **kwargs) -> Dict:
    base = dict()
    for __m in __object:
        if not (empty or __m): continue
        base = dict(base, **(diff_dict(__m, **base) if keep == "first" else __m))
    return base


def to_dict(__object: MappingData, **kwargs) -> Dict:
    if isinstance(__object, Dict): return __object
    elif is_records(__object, empty=True): return chain_dict(__object, keep="first")
    elif isinstance(__object, pd.DataFrame): return __object.to_dict("dict")
    else: return dict()


def keys_to_back(__m: Dict, __keys: _KT, **kwargs) -> Dict:
    __keys = cast_tuple(__keys)
    return dict(sorted(__m.items(), key=cmp_to_key(lambda x, y: -1 if y[0] in __keys else 0)))


def set_dict(__m: Dict, __keys: _KT, values: _VT, empty=True, inplace=True, **kwargs) -> Dict:
    if not inplace: __m = __m.copy()
    for __key, value in zip(cast_tuple(__keys), cast_tuple(values)):
        if value or empty: __m[__key] = value
    if not inplace: return __m


def drop_dict(__m: Dict, __keys: _KT, inplace=False, **kwargs) -> Dict:
    if not inplace: __m = __m.copy()
    for __key in cast_tuple(__keys): __m.pop(__key, None)
    if not inplace: return __m


def hier_get(__m: Dict, __path: _KT, default=None, apply: Optional[ApplyFunction]=None,
            __type: Optional[_type]=None, empty=True, null=True, **kwargs) -> _VT:
    try:
        for key in cast_tuple(__path):
            __m = __m[key]
        value = __apply(__m, apply, default, **kwargs) if apply else __m
        value = default if __type and not isinstance(value, __type) else value
        return value if (value or empty) and (value is not None or null) else default
    except: return default


def hier_set(__m: Dict, __path: _KT, value: _VT, empty=True, null=True, inplace=True, **kwargs) -> Dict:
    if not inplace: __m = __m.copy()
    for key in cast_tuple(__path)[:-1]:
        if (value or empty) and (value is not None or null): __m = __m[key]
    __m[__path[-1]] = value
    if not inplace: return __m


def hier_get_set(__m: Dict, __get_path: _KT, __set_path: _KT, default=None,
                apply: Optional[ApplyFunction]=None, __type: Optional[_type]=None,
                empty=True, null=True, inplace=True, **kwargs) -> Dict:
    value = hier_get(__m, __get_path, default, apply, __type, empty, null, **kwargs)
    if (value or empty) and (value is not None or null):
        return hier_set(__m, __set_path, value, inplace=inplace)
    else: return __m


def align_dict(__m: Dict[_KT,Sequence], default=None, empty=True, **kwargs) -> Dict[_KT,List]:
    if not __m: return dict()
    elif empty:
        count = max([len(values) for _, values in __m.items()])
        return {key: fill_array(values, count=count, value=default) for key, values in __m.items()}
    else:
        valid = __and(*[{i for i, value in enumerate(values) if value} for key, values in __m.items()])
        return {key: [values[i] for i in valid] for key, values in __m.items()}


def match_keywords(__m: Dict, __keys: _KT, include: Optional[Keyword]=list(),
                    exclude: Optional[Keyword]=list(), **kwargs) -> bool:
    if not (include or exclude): return True
    include, exclude, match = cast_tuple(include), cast_tuple(exclude), [False, False]
    for condition, keywords in enumerate([include, exclude]):
        pattern = re.compile('|'.join(map(re.escape, keywords)))
        for __key in cast_tuple(__keys):
            match[condition] |= bool(pattern.search(cast_str(__m[__key])))
    return (match[INCLUDE] or not include) and not (match[EXCLUDE] and exclude)


###################################################################
############################# Records #############################
###################################################################

def vloc(__r: List[Dict], __keys: _KT, default=None, value_only=False, match_type=False, **kwargs) -> Union[Records,List]:
    base = list()
    for __m in __r:
        values = kloc(__m, __keys, default=default, value_only=value_only, match_type=match_type)
        if default == "pass" and (values in (None, dict(), default)): continue
        else: base.append(values)
    return base


def isin_records(__r: Records, __keys: _KT, how="any", **kwargs) -> _bool:
    if not is_array(__keys):
        isin = [__keys in __m for __m in __r]
        return allin(isin) if how == "all" else any(isin)
    elif not __keys: return [True]*len(__r)
    else: return [isin_records(__r, __key, how=how) for __key in __keys if not is_array(__key)]


def to_records(__object: MappingData, **kwargs) -> Records:
    if is_records(__object, empty=True): return __object
    elif isinstance(__object, pd.DataFrame): return __object.to_dict("records")
    elif isinstance(__object, dict): return [__object]
    else: return list()


def apply_records(__r: List[Dict], __keys: Optional[_KT]=list(), __applyFunc: Optional[ApplyFunction]=list(),
                    all_keys=False, apply: Optional[ApplyFunction]=None, **context) -> Records:
    __r = __r.copy()
    if all_keys:
        if not isinstance(apply, Callable): return __r
        else: return [{__key:apply(__values) for __key, __values in __m.items()} for __m in __r]
    for idx, __m in enumerate(__r.copy()):
        for __key, applyFunc in map_context(__keys, __applyFunc, **context).items():
            if not isinstance(applyFunc, Callable): continue
            elif not __key: __r[idx] = applyFunc(__m)
            elif __key not in __m: continue
            else: __r[idx][__key] = applyFunc(__m[__key])
    return __r


def match_records(__r: List[Dict], __keys: Optional[_KT]=list(), __matchFunc: Optional[MatchFunction]=list(),
                    all_keys=False, **context) -> Sequence[bool]:
    __s = list()
    if all_keys:
        if not isinstance(__matchFunc, Callable): return [False]*len(__r)
        else: return [__and(map(__matchFunc, __m.values())) for __m in __r]
    context = map_context(__keys, __matchFunc, **context)
    context = {__key: __value for __key, __value in context.items() if isin_records(__r, __key)}
    for __m in __r:
        match = True
        for __key, matchFunc in context.items():
            if not isinstance(matchFunc, Callable): continue
            elif not __key: match &= matchFunc(__m)
            elif __key not in __m: continue
            else: match &= matchFunc(__m[__key])
        __s.append(__m)
    return __s


def between_records(__r: Records, __keys: Optional[_KT]=list(), __ranges: Optional[BetweenRange]=list(),
                    inclusive="both", if_null="drop", **context) -> Records:
    __s = list()
    for __m in __r:
        match = True
        for __key, __range in map_context(__keys, __ranges, **context).items():
            if __key in __m:
                if is_array(__range): match &= between(__m[__key], *__range[:2], inclusive=inclusive)
                elif isinstance(__range, dict): match &= between(__m[__key], **__range, inclusive=inclusive)
                else: raise ValueError("Between condition must be an iterable or a dictionary")
            elif if_null == "drop":
                match = False
                break
        if match: __s.append(__m)
    return __s


def sort_records(__r: Records, by: _KT, ascending=True, **kwargs) -> Records:
    return sorted(__r, key=lambda x: kloc(x, cast_tuple(by), value_only=True), reverse=(not ascending))


###################################################################
############################ DataFrame ############################
###################################################################

def cloc(df: pd.DataFrame, columns: IndexLabel, default=None, reorder=True, **kwargs) -> pd.DataFrame:
    if isinstance(columns, str):
        if columns in df: return df[[columns]]
        elif default == "keep": return pd.DataFrame([default]*len(df), columns=[columns])
        else: return pd.DataFrame()
    elif columns:
        if reorder: columns = [column for column in columns if column in df.columns]
        else: columns = [column for column in df.columns if column in columns]
        df = df[columns]
        if default == "keep": df = pd.concat([pd.DataFrame(columns=columns),df])
    return df


def to_dataframe(__object: MappingData, **kwargs) -> pd.DataFrame:
    if isinstance(__object, pd.DataFrame): return __object
    elif is_records(__object, empty=True): return pd.DataFrame(__object)
    elif isinstance(__object, dict): return pd.DataFrame([__object])
    else: return pd.DataFrame()


def apply_df(df: pd.DataFrame, __columns: Optional[IndexLabel]=list(), __applyFunc: Optional[ApplyFunction]=list(),
            all_cols=False, apply: Optional[ApplyFunction]=None, **context) -> pd.DataFrame:
    df = df.copy()
    if all_cols:
        if not isinstance(apply, Callable): return df
        for column in df.columns:
            df[column] = df[column].apply(apply)
        return df
    for column, applyFunc in map_context(__columns, __applyFunc, **context).items():
        if not (isinstance(column, str) and isinstance(applyFunc, Callable)): continue
        elif not column: df = df.apply(applyFunc, axis=1)
        elif column not in df: continue
        else: df[column] = df[column].apply(applyFunc)
    return df


def match_df(df: pd.DataFrame, __columns: Optional[IndexLabel]=list(),
            __matchFunc: Optional[MatchFunction]=list(), all_cols=False, **context) -> pd.DataFrame:
    df, match = df.copy(), pd.Series([True]*len(df), index=df.index)
    if all_cols:
        if not isinstance(__matchFunc, Callable): pd.Series([False]*len(df), index=df.index)
        for column in df.columns:
            match &= df[column].apply(__matchFunc)
        return df[match]
    for column, matchFunc in map_context(__columns, __matchFunc, **context).items():
        if not (isinstance(column, str) and isinstance(matchFunc, Callable)): continue
        elif not column: match &= df.apply(matchFunc, axis=1)
        elif column not in df: continue
        else: match &= df[column].apply(matchFunc)
    return df[match]


def concat_df(__object: Sequence[pd.DataFrame]) -> pd.DataFrame:
    __object = [df for df in __object if df_exists(df)]
    return pd.concat(__object) if __object else pd.DataFrame()


def between_df(df: pd.DataFrame, inclusive="both", if_null="drop", **context) -> pd.DataFrame:
    df, kwargs, default = df.copy(), {"inclusive":inclusive}, pd.Series([False]*len(df), index=df.index)
    for column, args in context.items():
        if column in df:
            if_na = default if if_null == "drop" else df[column].isna()
            if is_array(args):
                df = df[df[column].apply(lambda x: between(x, *args[:2], **kwargs))|if_na]
            elif isinstance(args, dict):
                df = df[df[column].apply(lambda x: between(x, **args, **kwargs))|if_na]
            else: raise ValueError("Between condition must be an iterable or a dictionary")
    return df


def merge_drop(left: pd.DataFrame, right: pd.DataFrame, drop="right", how="inner",
                on: Optional[IndexLabel]=None, **kwargs) -> pd.DataFrame:
    if df_empty(left) or df_empty(right):
        return right if drop == "left" else left
    key_in = lambda columns: not (set(cast_tuple(on)) & set(columns))
    if not on or key_in(left.columns) or key_in(right.columns):
        return right if drop == "left" else left
    duplicates = list((set(left.columns) & set(right.columns)) - set(cast_tuple(on)))
    if drop == "left": left = left.drop(columns=duplicates)
    elif drop == "right": right = right.drop(columns=duplicates)
    return left.merge(right, how=how, on=on, **kwargs)


def unroll_df(df: pd.DataFrame, columns: IndexLabel, values: _VT, **kwargs) -> pd.DataFrame:
    columns, values = cast_tuple(columns), cast_tuple(values)
    get_values = lambda row: [row[value] for value in values]
    len_values = lambda row: min(map(len, get_values(row)))
    unroll_row = lambda row: [[row[col]]*len_values(row) for col in columns]+get_values(row)
    map_subrow = lambda subrow: {key:value for key, value in zip(columns+values,subrow)}
    map_row = lambda row: pd.DataFrame([map_subrow(subrow) for subrow in zip(*unroll_row(row))])
    return pd.concat([map_row(row) for _,row in df.iterrows()])


def round_df(df: pd.DataFrame, columns: IndexLabel, trunc=2, **kwargs) -> pd.DataFrame:
    if not isinstance(trunc, int): return df
    roundFunc = lambda x: round(x,trunc) if isinstance(x,float) else x
    return apply_df(df, **{column:roundFunc for column in cast_tuple(columns)})


###################################################################
############################ Multitype ############################
###################################################################

def exists_one(*args, **kwargs) -> Any:
    for arg in args:
        if arg: return arg
    return args[-1]


def filter_exists(__object, null_only=False, **kwargs) -> Any:
    is_valid = pd.notna if null_only else exists
    if is_array(__object):
        return type(__object)([value for value in __object if is_valid(value)])
    elif isinstance(__object, Dict):
        return {key:value for key,value in __object.items() if is_valid(value)}
    elif isinstance(__object, pd.DataFrame):
        return __object.dropna(axis=1, how="any")
    elif is_valid(__object): return __object


def convert_data(data: Data, returnType: Optional[TypeHint]=None, **kwargs) -> Data:
    if not returnType: return data
    elif is_records_type(returnType): return to_records(data)
    elif is_dataframe_type(returnType): return to_dataframe(data)
    elif is_dict_type(returnType): return to_dict(data)
    elif is_list_type(returnType): return cast_list(data)
    else: return data


def chain_exists(data: Data, returnType: Optional[TypeHint]=None, **kwargs) -> Data:
    if is_dfarray(data): data = concat_df(data)
    elif is_2darray(data): data = chain_array(data, empty=False)
    else: data = filter_exists(data)
    return convert_data(data, returnType)


def filter_data(data: Data, filter: Optional[Union[_KT,Index]]=list(), default=None,
                returnType: Optional[TypeHint]=None, **kwargs) -> Data:
    if not filter: return convert_data(data, returnType)
    filter = cast_tuple(filter)
    if is_records(data): data = vloc(data, __keys=filter, default=default)
    elif isinstance(data, pd.DataFrame): data = cloc(data, columns=filter, default=default)
    elif isinstance(data, Dict): data = kloc(data, __keys=filter, default=default)
    elif isinstance(data, List): data = iloc(data, index=filter, default=default)
    return convert_data(data, returnType)


def apply_data(data: Data, __keys: Optional[Union[_KT,Index]]=list(), __applyFunc: Optional[ApplyFunction]=list(),
                all_keys=False, apply: Optional[ApplyFunction]=None, returnType: Optional[TypeHint]=None, **context) -> Data:
    if is_records(data): data = apply_records(data, __keys, __applyFunc, all_keys=all_keys, apply=apply, **context)
    elif isinstance(data, pd.DataFrame): data = apply_df(data, __keys, __applyFunc, all_cols=all_keys, apply=apply, **context)
    elif isinstance(data, Dict): data = apply_dict(data, __keys, __applyFunc, all_keys=all_keys, apply=apply, **context)
    elif isinstance(data, List): data = apply_array(data, __keys, __applyFunc, all_indices=all_keys, apply=apply, **context)
    return convert_data(data, returnType)


def match_data(data: Data, __keys: Optional[Union[_KT,Index]]=list(), __matchFunc: Optional[MatchFunction]=list(),
                all_keys=False, returnType: Optional[TypeHint]=None, **context) -> Data:
    if is_records(data): data = is_records_type(data, __keys, __matchFunc, all_keys=all_keys, **context)
    elif isinstance(data, pd.DataFrame): data = match_df(data, __keys, __matchFunc, all_cols=all_keys, **context)
    elif isinstance(data, Dict): data = match_dict(data, __keys, __matchFunc, all_keys=all_keys, **context)
    elif isinstance(data, List): data = match_array(data, __keys, __matchFunc, all_indices=all_keys, **context)
    return convert_data(data, returnType)


def between_data(data: TabularData, inclusive="both", if_null="drop",
                returnType: Optional[TypeHint]=None, **kwargs) -> TabularData:
    if is_records(data): data = between_records(data, inclusive=inclusive, if_null=if_null, **kwargs)
    elif isinstance(data, pd.DataFrame): data = between_df(data, inclusive=inclusive, if_null=if_null)
    return convert_data(data, returnType)


def sort_values(data: TabularData, by: _KT, ascending: _bool=True,
                returnType: Optional[TypeHint]=None, **kwargs) -> TabularData:
    if is_records(data):
        ascending = bool(iloc(ascending, 0)) if isinstance(ascending, Sequence) else ascending
        data = sort_records(data, by=by, ascending=ascending)
    elif isinstance(data, pd.DataFrame): data = data.sort_values(by, ascending=ascending)
    return convert_data(data, returnType)



def exists_context(**context):
    return {key: value for key, value in context.items() if value or isinstance(value, (bool,float,int))}


CONFIG_CONTEXT = lambda asyncio=False, operation=str(), host=str(), where=str(), which=str(), initTime=None, \
                        session=None, semaphore=None, filter=list(), filterContext=None, \
                        contextArgs=list(), contextQuery=list(), contextUnit=1, contextPeriod=str(), \
                        startDate=None, endDate=None, returnType=str(), \
                        logName=str(), logLevel="WARN", logFile=str(), logJson=False, logger=None, \
                        logErrors=False, errorArgs=tuple(), errorKwargs=tuple(), errors=dict(), \
                        delay=1., numTasks=100, progress=True, debug=False, message=str(), rename=dict(), \
                        apiRedirect=False, reidrectUnit=1, redirectErrors=False, localSave=False, extraSave=False, \
                        queryKey=str(), querySheet=str(), queryFields=list(), queryString=list(), queryArray=list(), \
                        authType=None, idKey=None, pwKey=None, extraKeys=None, dependencies=list(), \
                        data=None, results=None, crawler=None, prefix=str(), self_var=True, **context: \
    dict(exists_context(
        filter = filter,
        startDate = startDate,
        endDate = endDate,
        logName = logName,
        logLevel = logLevel,
        logFile = logFile,
        logErrors = logErrors,
        errorArgs = errorArgs,
        errorKwargs = errorKwargs,
        delay = delay,
        numTasks = numTasks,
        progress = progress,
        debug = debug,
        message = message,
        rename = rename,
        apiRedirect = apiRedirect,
        reidrectUnit = reidrectUnit,
        redirectErrors = redirectErrors,
        ), **context)


ENC_CONTEXT = lambda encryptedKey=str(), decryptedKey=dict(), cookies=str(), userid=str(), passwd=str(), \
                    **context: \
    dict(exists_context(
        encryptedKey = encryptedKey,
        cookies = cookies,
        ), **context)


UPLOAD_CONTEXT = lambda gsKey=str(), gsSheet=str(), gsMode="append", gsBaseSheet=str(), gsRange=str(), \
                    gbqPid=str(), gbqTable=str(), gbqMode="append", gbqSchema=None, gbqProgress=True, \
                    gbqPartition=str(), gbqPartitionBy="auto", gbqReauth=False, \
                    **context: context


GS_CONTEXT = lambda key=str(), sheet=str(), mode="append", base_sheet=str(), cell=str(), clear=False, **context: context


GBQ_CONTEXT = lambda table=str(), project_id=str(), mode="append", schema=None, progress=True, \
                    partition=str(), partition_by="auto", reauth=False, **context: context


PROXY_CONTEXT = lambda **context: CONFIG_CONTEXT(**UPLOAD_CONTEXT(**context))


REDIRECT_CONTEXT = lambda apiRedirect=False, redirectUnit=(1,), redirectErrors=False, logFile=str(), **context: \
    CONFIG_CONTEXT(**ENC_CONTEXT(**UPLOAD_CONTEXT(**context)))


GCP_CONTEXT = lambda **context: GS_CONTEXT(**GBQ_CONTEXT(**context))



from .cast import cast_int, cast_datetime, cast_date, get_timezone
from .map import re_get
from .types import TypeHint, DateNumeric, DateFormat, Timedelta, Timezone
from .types import is_type, is_str_type, is_timestamp_type, INTEGER_TYPES

from typing import List, Optional, Union
from dateutil.parser import parse as dateparse
from pandas.tseries.offsets import BDay
from pytz import timezone
import datetime as dt
import numpy as np
import pandas as pd
import re

KST = "Asia/Seoul"
NOW = dt.datetime.now(timezone(KST)).replace(microsecond=0, tzinfo=None)
TODAY = NOW.replace(hour=0, minute=0, second=0, microsecond=0)

DATEPART = ["SECOND", "MINUTE", "HOUR", "DAY", "MONTH", "YEAR"]
WEEKDAY = ["월", "화", "수", "목", "금", "토", "일"]

DATETIME_FORMAT = {
    "date": "%Y-%m-%d",
    "datetime": "%Y-%m-%d %H:%M:%S",
    "datetime_ms": "%Y-%m-%d %H:%M:%S.%f",
    "timezone": "%Y-%m-%d %H:%M:%S%z",
    "timezone_ms": "%Y-%m-%d %H:%M:%S.%f%z",
    "js": "%Y-%m-%dT%H:%M:%S.%f%z",
}

DATETIME_PATTERN = {
    "date": r"^\d{4}-\d{2}-\d{2}$",
    "datetime": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$",
    "datetime_ms": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+$",
    "timezone": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\+\d{2}:\d{2}$",
    "timezone_ms": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+\+\d{2}:\d{2}$",
    "js": r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}$",
}

COMMON_DATE_PATTERN = r"\d{4}-\d{2}-\d{2}"
COMMON_DATETIME_PATTERN = r"^(?:\d{4}-\d{2}-\d{2}(?: \d{2}:\d{2}:\d{2}(?:\.\d{1,6})?(?:[+-]\d{4})?)?|\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?(?:[+-]\d{4})?)$"

PANDAS_FREQUENCY = {
    "B": "business day frequency",
    "C": "custom business day frequency",
    "D": "calendar day frequency",
    "W": "weekly frequency",
    "M": "month end frequency",
    "SM": "semi-month end frequency (15th and end of month)",
    "BM": "business month end frequency",
    "CBM": "custom business month end frequency",
    "MS": "month start frequency",
    "SMS": "semi-month start frequency (1st and 15th)",
    "BMS": "business month start frequency",
    "CBMS": "custom business month start frequency",
    "Q": "quarter end frequency",
    "BQ": "business quarter end frequency",
    "QS": "quarter start frequency",
    "BQS": "business quarter start frequency",
    "A": "year end frequency",
    "Y": "year end frequency",
    "BA": "business year end frequency",
    "BY": "business year end frequency",
    "AS": "year start frequency",
    "YS": "year start frequency",
    "BAS": "business year start frequency",
    "BYS": "business year start frequency",
    "BH": "business hour frequency",
    "H": "hourly frequency",
    "T": "minutely frequency",
    "min": "minutely frequency",
    "S": "secondly frequency",
    "L": "milliseconds",
    "ms": "milliseconds",
    "U": "microseconds",
    "us": "microseconds",
    "N": "nanoseconds"
}


strptimekr = lambda date_string: cast_datetime(date_string, tzinfo=KST, droptz=True)
strpdatekr = lambda date_string: cast_date(strptimekr(date_string))


def trunc_datetime(__datetime: dt.datetime, part=str(), **kwargs) -> dt.datetime:
    if part not in DATEPART: return __datetime
    index = DATEPART.index(part)
    if index >= 0: __datetime = __datetime.replace(microsecond=0)
    if index >= 1: __datetime = __datetime.replace(second=0)
    if index >= 2: __datetime = __datetime.replace(minute=0)
    if index >= 3: __datetime = __datetime.replace(hour=0)
    if index >= 4: __datetime = __datetime.replace(day=1)
    if index >= 5: __datetime = __datetime.replace(month=1)
    return __datetime


def now(__format=str(), days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0,
        hours=0, weeks=0, tzinfo=KST, droptz=True, datetimePart="SECOND", **kwargs) -> Union[dt.datetime,str]:
    delta = dt.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)
    __datetime = dt.datetime.now(get_timezone(tzinfo)) - delta
    if droptz: __datetime = __datetime.replace(tzinfo=None)
    if datetimePart: __datetime = trunc_datetime(__datetime, datetimePart)
    return __datetime.strftime(__format) if __format else __datetime


def today(__format=str(), days=0, weeks=0, tzinfo=KST, **kwargs) -> Union[dt.date,str]:
    if __format: return now(__format, days=days, weeks=weeks, tzinfo=tzinfo, datetimePart="DAY")
    else: return now(days=days, weeks=weeks, tzinfo=tzinfo, datetimePart="DAY").date()


def get_datetime(__object: Optional[DateFormat]=None, days=0, seconds=0, microseconds=0, milliseconds=0,
                minutes=0, hours=0, weeks=0, tzinfo=None, astimezone=None, droptz=False, datetimePart="SECOND",
                **kwargs) -> dt.datetime:
    context = dict(tzinfo=tzinfo, astimezone=astimezone, droptz=droptz)
    __datetime = cast_datetime(__object, **context)
    __datetime = __datetime if isinstance(__datetime, dt.datetime) else now(days=cast_int(__object), **context)
    if isinstance(__datetime, dt.datetime):
        __datetime = __datetime - dt.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)
        return trunc_datetime(__datetime, datetimePart) if datetimePart else __datetime


def get_time(__object: Optional[DateFormat]=None, days=0, seconds=0, microseconds=0, milliseconds=0,
            minutes=0, hours=0, weeks=0, tzinfo=None, astimezone=None, datetimePart="SECOND",
            **kwargs) -> dt.time:
    __datetime = get_datetime(**locals())
    if isinstance(__datetime, dt.datetime):
        return __datetime.time()


def get_timestamp(__object: Optional[DateFormat]=None, days=0, seconds=0, microseconds=0, milliseconds=0,
                minutes=0, hours=0, weeks=0, tzinfo=None, astimezone=None, droptz=False, datetimePart="SECOND",
                tsUnit="ms", **kwargs) -> int:
    __datetime = get_datetime(**locals())
    if isinstance(__datetime, dt.datetime):
        return int(__datetime.timestamp()*(1000 if tsUnit == "ms" else 1))


def get_date(__object: Optional[DateFormat]=None, days=0, weeks=0, tzinfo=None, **kwargs) -> dt.date:
    __date = cast_date(__object, tzinfo=tzinfo)
    __date = __date if isinstance(__date, dt.date) else today(days=cast_int(__object), tzinfo=tzinfo)
    if isinstance(__date, dt.date):
        return __date - dt.timedelta(days=days, weeks=weeks)


def get_busdate(__object: Optional[DateFormat]=None, days=0, weeks=0, tzinfo=None, **kwargs) -> dt.date:
    __date = get_date(**locals())
    if isinstance(__date, dt.date):
        return __date if np.is_busday(__date) else (__date-BDay(1)).date()


def is_datetime_format(__object: DateFormat, strict=False, **kwargs) -> bool:
    if isinstance(__object, str) and re.fullmatch(COMMON_DATETIME_PATTERN, __object):
        try: return bool(dateparse(__object)) if strict else True
        except: return False
    else: return False


def get_datetime_format(__object: DateFormat, strict=False, **kwargs) -> str:
    if is_datetime_format(__object, strict=strict):
        return "date" if re.fullmatch(COMMON_DATE_PATTERN, __object) else "datetime"


def cast_datetime_format(__object: DateFormat, default=None, strict=False, **kwargs) -> DateNumeric:
    format = get_datetime_format(__object, strict=strict)
    if format == "date": return cast_date(__object, default=default)
    elif format == "datetime": return cast_datetime(__object, default=default)
    else: return default


def is_pandas_frequency(__object: Timedelta, **kwargs) -> bool:
    if isinstance(__object, dt.timedelta): return True
    elif isinstance(__object, str): return any(map(lambda freq: __object.endswith(freq), PANDAS_FREQUENCY.keys()))
    else: return False


def get_date_range(startDate: Optional[DateFormat]=None, endDate: Optional[DateFormat]=None,
                    periods: Optional[int]=None, interval: Timedelta="D",
                    tzinfo: Optional[Timezone]=None, **kwargs) -> List[dt.date]:
    if sum(map(bool, (startDate, endDate, periods, interval))) < 3: return list()
    if isinstance(interval, int):
        interval = dt.timedelta(days=interval)
    if not is_pandas_frequency(interval): return list()
    else: return [date.date() for date in pd.date_range(startDate, endDate, periods, interval, tzinfo)]


def set_datetime(__datetime: dt.datetime, __type: TypeHint=str,
                __format="%Y-%m-%d %H:%M:%S", tsUnit="ms", **kwargs) -> Union[str,int]:
    if __format: return __datetime.strftime(DATETIME_FORMAT.get(__format,__format))
    elif is_str_type(__type): return str(__datetime)
    elif is_timestamp_type(__type): return get_timestamp(__datetime, tsUnit=tsUnit)
    else: return


def set_date(__date: dt.date, __type: TypeHint=str,
            __format="%Y-%m-%d", **kwargs) -> Union[str,int]:
    if __format: return __date.strftime(DATETIME_FORMAT.get(__format,__format))
    elif is_str_type(__type): return str(__date)
    elif is_type(__type, INTEGER_TYPES+["ordinal"]): return __date.toordinal()
    else: return


def relative_strptime(__date_string: str, **kwargs) -> dt.datetime:
    str2int = lambda pattern, string: cast_int(re_get(pattern, string))
    if re.search("(\d+)초 전", __date_string):
        return now(seconds=str2int("(\d+)초 전", __date_string))
    elif re.search("(\d+)분 전", __date_string):
        return now(minutes=str2int("(\d+)분 전", __date_string))
    elif re.search("(\d+)시간 전", __date_string):
        return now(hours=str2int("(\d+)시간 전", __date_string))
    elif re.search("(\d+)일 전", __date_string):
        return today(days=str2int("(\d+)일 전", __date_string))
    elif "어제" in __date_string:
        return today(days=1)
    else: return cast_datetime(__date_string)


def strfweek(__date: dt.date, **kwargs) -> str:
    cur_year, weeknum, _ = __date.isocalendar()
    year, first_week, first_day = __date.replace(day=1).isocalendar()
    if cur_year != year:
        return f"{cur_year}년{__date.month}월{weeknum}주차"
    monthly_week = weeknum - first_week + int(first_day < 4)
    return f"{year}년{__date.month}월{monthly_week}주차"



from .cast import cast_datetime, cast_date
from .map import re_get
from .types import JsonData, IndexLabel

from typing import Dict, List, Union
from ast import literal_eval
from bs4 import BeautifulSoup
from bs4.element import Tag
from http.cookies import SimpleCookie
from requests.cookies import RequestsCookieJar
import datetime as dt
from urllib.parse import quote, urlparse
import json
import re

import warnings
warnings.filterwarnings("ignore", category=UserWarning, module='bs4')


clean_html = lambda html: BeautifulSoup(str(html), "lxml").text
clean_tag = lambda source: re.sub("<[^>]*>", "", str(source))


def parse_cookies(cookies: Union[RequestsCookieJar,SimpleCookie]) -> str:
    return '; '.join([str(key)+"="+str(value) for key,value in cookies.items()])


def parse_parth(url: str) -> str:
    return re.sub(urlparse(url).path+'$','',url)


def parse_origin(url: str) -> str:
    return re_get(f"(.*)(?={urlparse(url).path})", url) if urlparse(url).path else url


def encode_cookies(cookies: Union[Dict,str], **kwargs) -> str:
    return '; '.join(
    [(parse_cookies(data) if isinstance(data, Dict) else str(data)) for data in [cookies,kwargs] if data])


def encode_object(__object: str) -> str:
    quote(str(__object).replace('\'','\"'))


class LazyDecoder(json.JSONDecoder):
    def decode(s, **kwargs):
        regex_replacements = [
            (re.compile(r'([^\\])\\([^\\])'), r'\1\\\\\2'),
            (re.compile(r',(\s*])'), r'\1'),
        ]
        for regex, replacement in regex_replacements:
            s = regex.sub(replacement, s)
        return super().decode(s, **kwargs)


def validate_json(data: JsonData, __path: IndexLabel, default=dict(), **kwargs) -> JsonData:
    __m = data.copy()
    try:
        for key in __path:
            __m = __m[key]
            if isinstance(__m, str):
                try: __m = json.loads(__m)
                except json.JSONDecodeError: return json.loads(__m, cls=LazyDecoder)
        return __m
    except: return default


def parse_invalid_json(raw_json: str, key: str, value_type="dict", **kwargs) -> JsonData:
    rep_bool = lambda s: str(s).replace("null","None").replace("true","True").replace("false","False")
    try:
        if value_type == "dict" and re.search("\""+key+"\":\{[^\}]*\}+",raw_json):
            return literal_eval(rep_bool("{"+re.search("\""+key+"\":\{[^\}]*\}+",raw_json).group()+"}"))
        elif value_type == "any" and re.search(f"(?<=\"{key}\":)"+"([^,}])+(?=[,}])",raw_json):
            return literal_eval(rep_bool(re.search(f"(?<=\"{key}\":)"+"([^,}])+(?=[,}])",raw_json).group()))
    except:
        return dict()


def select_text(source: Tag, selector: str, pattern='\n', sub=' ', multiple=False, **kwargs) -> Union[str,List[str]]:
    try:
        if multiple: return [re.sub(pattern, sub, select.text).strip() for select in source.select(selector)]
        else: return re.sub(pattern, sub, source.select_one(selector).text).strip()
    except (AttributeError, IndexError, TypeError):
        return list() if multiple else str()


def select_attr(source: Tag, selector: str, key: str, default=None, multiple=False, **kwargs) -> Union[str,List[str]]:
    try:
        if multiple: return [select.attrs.get(key,default).strip() for select in source.select(selector)]
        else: return source.select_one(selector).attrs.get(key,default).strip()
    except (AttributeError, IndexError, TypeError):
        return list() if multiple else str()


def select_datetime(source: Tag, selector: str, default=None, multiple=False, **kwargs) -> dt.datetime:
    if multiple:
        return [cast_datetime(text, default) for text in select_text(source, selector, multiple=True)]
    else:
        return cast_datetime(select_text(source, selector, multiple=False), default)


def select_date(source: Tag, selector: str, default=None, multiple=False, **kwargs) -> dt.datetime:
    if multiple:
        return [cast_date(text, default) for text in select_text(source, selector, multiple=True)]
    else:
        return cast_date(select_text(source, selector, multiple=False), default)


def match_class(source: Tag, class_name: str, **kwargs) -> bool:
    try:
        class_list = source.attrs.get("class")
        if isinstance(class_list, List):
            return class_name in class_list
        return class_name == class_list
    except (AttributeError, TypeError):
        return False



from .cast import cast_str, cast_float, cast_int, to_datetime
from .date import is_datetime_format, get_datetime, get_timestamp, get_time, get_date, DATEPART
from .map import iloc, cloc, apply_df, apply_data, df_exists, df_empty
from .types import RenameDict, TabularData, TypeHint, is_records

from google.oauth2 import service_account
from google.oauth2.service_account import IDTokenCredentials
from google.auth.transport.requests import AuthorizedSession

from gspread.worksheet import Worksheet
import gspread

from pandas_gbq.gbq import InvalidSchema
import pandas_gbq

from typing import Any, Dict, List, Optional, Sequence, Union
from tqdm.auto import tqdm
import datetime as dt
import functools
import json
import os
import pandas as pd
import requests

Account = Union[Dict[str,str], str]
PostData = Union[Dict[str,Any],str]
Datetime = Union[dt.datetime, dt.date]

NumericiseIgnore = Union[bool, Sequence[int]]

BigQuerySchema = List[Dict[str,str]]
SchemaSequence = Union[BigQuerySchema, Sequence[BigQuerySchema]]

ENV_PATH = "env/"
GCLOUD_ACCOUNT = ENV_PATH+"gcloud.json"
GCLOUD_DATA = ENV_PATH+"data.json"


INVALID_SCHEMA_MSG = "Please verify that the structure and data types in the DataFrame match the schema of the destination table."
INVALID_UPSERT_KEY_MSG = "Please verify that the key exists and is in both DataFrame objects."
BIGQUERY_PARTITION_MSG = "Uploading partitioned data to Google BigQuery."


###################################################################
###################### Google Authorization #######################
###################################################################

def read_json(file: str) -> Dict:
    if os.path.exists(file):
        with open(file, "r", encoding="utf-8") as f:
            return json.loads(f.read())
    else: return dict()


def read_gcloud(file=str(), **kwargs) -> Account:
    return read_json(str(file) if str(file).endswith(".json") else GCLOUD_ACCOUNT)


def read_data(file=str(), **kwargs) -> PostData:
    return read_json(str(file) if str(file).endswith(".json") else GCLOUD_DATA)


def fetch_gcloud_credentials(audience=str(), account: Account=dict(), **kwargs) -> IDTokenCredentials:
    account = account if account and isinstance(account, dict) else read_gcloud(account)
    audience = audience if audience else account.get("audience", str())
    return service_account.IDTokenCredentials.from_service_account_info(account, target_audience=audience)


def fetch_gcloud_authorization(audience=str(), account: Account=dict(), **kwargs) -> str:
    credentials = fetch_gcloud_credentials(audience, account)
    auth_session = AuthorizedSession(credentials)
    auth_session.get(audience)
    return "Bearer "+credentials.token


def gcloud_authorized(func):
    @functools.wraps(func)
    def wrapper(*args, audience=str(), authorization=str(), account: Account=dict(), **kwargs):
        if not authorization:
            authorization = fetch_gcloud_authorization(audience, account, **kwargs)
        return func(*args, audience=audience, authorization=authorization, account=account, **kwargs)
    return wrapper


@gcloud_authorized
def request_gcloud(audience: str, authorization: str, data: Optional[PostData]=dict(),
                    operation=str(), **kwargs) -> requests.Response:
    data = data if data and isinstance(data, dict) else read_data(str(data)).get(operation, dict())
    return requests.post(audience, json=data, headers={"Authorization":authorization})


###################################################################
###################### Google Spread Sheets #######################
###################################################################

def to_excel_date(date: Datetime) -> int:
        offset = 693594
        days = date.toordinal() - offset
        if isinstance(date, dt.datetime):
            seconds = (date.hour*60*60 + date.minute*60 + date.second)/(24*60*60)
            return days + seconds
        return days


def validate_excel_format(data: TabularData, action="download", returnType: Optional[TypeHint]=None, **kwargs) -> TabularData:
    if action == "download":
        cast_boolean = lambda x: {"TRUE":True, "FALSE":False}.get(x, x)
        applyFunc = lambda x: to_datetime(x, __type="auto") if is_datetime_format(x) else cast_boolean(x)
    elif action == "upload":
        applyFunc = lambda x: to_excel_date(x) if isinstance(dt.date) else x
    return apply_data(data, __applyFunc=applyFunc, all_keys=True, returnType=returnType)


def load_gspread(key: str, sheet: str, account: Account=dict(), **kwargs) -> Worksheet:
    account = account if account and isinstance(account, dict) else read_gcloud(account)
    gs_acc = gspread.service_account_from_dict(account)
    gs = gs_acc.open_by_key(key)
    return gs.worksheet(sheet)


def read_gspread(key: str, sheet: str, account: Account=dict(), gs: Optional[Worksheet]=None,
                head=1, headers=None, if_null=str(), numericise_ignore: NumericiseIgnore=list(),
                rename: RenameDict=dict(), **kwargs) -> pd.DataFrame:
    gs = gs if gs else load_gspread(key, sheet, account)
    if isinstance(numericise_ignore, bool): numericise_ignore = ["all"] if numericise_ignore else list()
    params = dict(head=head, default_blank=if_null, expected_headers=headers, numericise_ignore=numericise_ignore)
    data = pd.DataFrame(gs.get_all_records(**params)).rename(columns=rename)
    return validate_excel_format(data, action="download", returnType="dataframe")


def clear_gspead(key: str, sheet: str, account: Account=dict(),
                gs: Optional[Worksheet]=None, include_header=False, **kwargs):
    gs = gs if gs else load_gspread(key, sheet, account)
    if include_header: return gs.clear()
    last_row = len(gs.get_all_records())+1
    if last_row > 2: gs.delete_rows(3, last_row)


def update_gspread(key: str, sheet: str, data: TabularData, account: Account=dict(),
                    gs: Optional[Worksheet]=None, col='A', row=0, cell=str(), clear=False, clear_header=False, **kwargs):
    gs = gs if gs else load_gspread(key, sheet, account)
    records = validate_excel_format(data, action="upload", returnType="records")
    if not records: return
    values = [[value if pd.notna(value) else None for value in record.values()] for record in records]
    if clear: clear_gspead(gs=gs, include_header=clear_header)
    cell = cell if cell else (col+str(row if row else len(gs.get_all_records())+2))
    gs.update(cell, values)


###################################################################
######################### Google Bigquery #########################
###################################################################

BIGQUERY_TYPES = lambda __type, fillna=False: {
    "STRING": lambda x: x if isinstance(x, str) else cast_str(x, default=(str() if fillna else None), match=pd.notna),
    "BYTES": lambda x: x if isinstance(x, bytes) else None,
    "INTEGER": lambda x: x if isinstance(x, int) else cast_int(x, default=(0 if fillna else None)),
    "FLOAT": lambda x: x if isinstance(x, float) else cast_float(x, default=(0. if fillna else None)),
    "NUMERIC": lambda x: x if isinstance(x, int) else cast_int(x, default=(0 if fillna else None)),
    "BIGNUMERIC": lambda x: x if isinstance(x, int) else cast_int(x, default=(0 if fillna else None)),
    "BOOLEAN": lambda x: x if isinstance(x, bool) else bool(x),
    "TIMESTAMP": lambda x: x if isinstance(x, int) else get_timestamp(x, default=(0 if fillna else None), tsUnit="ms"),
    "DATE": lambda x: x if isinstance(x, dt.date) else get_date(x, default=(0 if fillna else None)),
    "TIME": lambda x: x if isinstance(x, dt.time) else get_time(x, default=(0 if fillna else None)),
    "DATETIME": lambda x: x if isinstance(x, dt.datetime) else get_datetime(x, default=(0 if fillna else None)),
}.get(__type, lambda x: x)


def gbq_authorized(func):
    @functools.wraps(func)
    def wrapper(*args, audience=str(), credentials: Optional[IDTokenCredentials]=None, account=dict(), reauth=False, **kwargs):
        if reauth and not credentials:
            credentials = fetch_gcloud_credentials(audience, account, **kwargs)
        return func(*args, audience=audience, credentials=credentials, account=account, reauth=reauth, **kwargs)
    return wrapper


def validate_schema(data: pd.DataFrame, schema: Optional[BigQuerySchema]=None, fillna=False, **kwargs) -> pd.DataFrame:
    if not (schema and is_records(schema)): return data
    context = {field["name"]:BIGQUERY_TYPES(field["type"], fillna) for field in schema}
    data = apply_df(cloc(data, list(context.keys()), default="pass"), **context)
    if df_exists(data, null=False): return data
    else: raise InvalidSchema(INVALID_SCHEMA_MSG, local_schema=data.dtypes.to_frame().to_dict()[0], remote_schema=schema)


@gbq_authorized
def read_gbq(query: str, project_id: str, reauth=False, credentials: Optional[IDTokenCredentials]=None, **kwargs) -> pd.DataFrame:
    return pd.read_gbq(query, project_id, reauth=reauth, credentials=credentials)


@gbq_authorized
def to_gbq(table: str, project_id: str, data: pd.DataFrame, reauth=False, if_exists="append",
            schema: Optional[BigQuerySchema]=None, progress=True, validate=True, fillna=False,
            partition=str(), partition_by="auto", credentials: Optional[IDTokenCredentials]=None, **kwargs):
    if validate: data = validate_schema(data, schema, fillna=fillna)
    context = dict(reauth=reauth, if_exists=if_exists, credentials=credentials)
    if partition:
        return to_gbq_partition(table, project_id, data, schema=schema, progress=progress, validate=False,
                                partition=partition, partition_by=partition_by, **context)
    else: data.to_gbq(table, project_id, table_schema=schema, progress_bar=progress, **context)


@gbq_authorized
def to_gbq_partition(table: str, project_id: str, data: pd.DataFrame, reauth=False, if_exists="append",
                    schema: Optional[BigQuerySchema]=None, progress=True, validate=True, fillna=False,
                    partition=str(), partition_by="auto", credentials: Optional[IDTokenCredentials]=None, **kwargs):
    if validate: data = validate_schema(data, schema, fillna=fillna)
    context = dict(destination_table=table, project_id=project_id, reauth=reauth, if_exists=if_exists,
                    table_schema=schema, progress_bar=progress, credentials=credentials)
    if partition not in data: data.to_gbq(**context)
    elif partition_by in DATEPART+["DATE"]:
        set_partition = (lambda x: get_datetime(x, datetimePart=partition_by)) if partition_by in DATEPART else get_date
        data["_PARTITIONTIME"] = data[partition].apply(set_partition)
        for date in tqdm(sorted(data["_PARTITIONTIME"].unique()), desc=BIGQUERY_PARTITION_MSG):
            data[data["_PARTITIONTIME"]==date].drop(columns="_PARTITIONTIME").to_gbq(**context)
    else:
        for part in tqdm(sorted(data[partition].unique()), desc=BIGQUERY_PARTITION_MSG):
            data[data[partition]==part].to_gbq(**context)


def validate_upsert_key(data: pd.DataFrame, base: pd.DataFrame, key=str(),
                        schema: Optional[BigQuerySchema]=None, index=0, **kwargs) -> str:
    if not (key and isinstance(key, str)) and (schema and is_records(schema)):
        keys = [field.get("name") for field in schema if field.get("mode") == "REQUIRED"]
        key = iloc(keys, index, default=str())
    if key and key in data and key in base: return key
    else: raise ValueError(INVALID_UPSERT_KEY_MSG)


@gbq_authorized
def upsert_gbq(table: str, project_id: str, data: pd.DataFrame, base: Optional[pd.DataFrame]=None, key=str(),
                reauth=False, if_exists="replace", schema: Optional[BigQuerySchema]=None, progress=True,
                validate=True, fillna=False, credentials: Optional[IDTokenCredentials]=None, **kwargs):
    if validate: data = validate_schema(data, schema, fillna=fillna)
    context = dict(reauth=reauth, credentials=credentials)
    if df_empty(base): base = read_gbq(table, project_id, **context)
    key = validate_upsert_key(data, base, key, schema)
    data = data.set_index(key).combine_first(base.set_index(key)).reset_index()
    data.to_gbq(table, project_id, if_exists=if_exists, table_schema=schema, progress_bar=progress, **context)



from typing import Callable, Sequence, Tuple, Type, Union, TypeVar, Hashable
from typing import Any, Dict, List, Set, get_type_hints, get_origin, get_args

from datetime import datetime, date, time, timedelta
from pandas import DataFrame
from pytz import BaseTzInfo


_KT = TypeVar("_KT", Hashable, Sequence[Hashable])
_VT = TypeVar("_VT", Any, Sequence[Any])
_bool = TypeVar("_bool", bool, Sequence[bool])
_type = TypeVar("_type", Type, Sequence[Type])

Comparable = TypeVar("Comparable")
ClassInstance = TypeVar("ClassInstance")

Arugments = Tuple[Any]
ArgsMapper = Callable[[Arugments],Arugments]

Context = Dict[Any,Any]
ContextMapper = Callable[[Context],Context]

ObjectSequence = Union[List, Tuple]
OBJECT_SEQUENCE = (List, Tuple)

IndexedSequence = Union[List, Tuple]
INDEXED_SEQUENCE = (List, Tuple)

SequenceSet = Union[ObjectSequence, Set]
SEQUENCE_SET = (List, Set, Tuple)

NestedSequence = Sequence[Sequence]
NestedDict = Dict[_KT,Dict]
RenameDict = Dict[str,str]
Records = List[Dict]

NestedData = Union[NestedSequence, NestedDict]
TabularData = Union[Records, DataFrame]
MappingData = Union[Records, DataFrame, Dict]
Data = Union[Records, DataFrame, Dict, List, NestedSequence, NestedDict]

JsonData = Union[Dict, List]
RedirectData = Union[Records, Dict[str,Records], JsonData]

LogMessage = Dict[str,str]

Index = Union[int, Sequence[int]]
IndexLabel = Union[Hashable, Sequence[Hashable]]

Column = Union[str, Sequence[str]]
Keyword = Union[str, Sequence[str]]
Shape = Union[int, Sequence[int]]
Unit = Union[int, Sequence[int]]

Timestamp = Union[float, int]
DateNumeric = Union[datetime, date, time, float, int]
DateFormat = Union[datetime, date, time, float, int, str]

DateQuery = Dict[str,datetime]
Timedelta = Union[timedelta, str, int]
Timezone= Union[BaseTzInfo, str]

PatternStr = str
ApplyFunction = Union[Callable[[Any],Any], Sequence[Callable[[Any],Any]]]
MatchFunction = Union[Callable[[Any],bool], Sequence[Callable[[Any],bool]]]
BetweenRange = Union[Sequence[Tuple], Sequence[Dict]]

CastError = (ValueError, TypeError)


def init_origin(__object, default=None, **kawrgs) -> Any:
    try:
        __type = get_origin(__object)
        if __type == Union: return init_origin(get_args(__object)[0])
        elif __type: return __type()
        elif isinstance(__object, Type): return __object()
        elif __object is None: return default
        return init_origin(get_type_hints(__object).get("return") if isinstance(__object, Callable) else type(__object))
    except: return default


###################################################################
############################ Type Check ###########################
###################################################################

TypeHint = Union[Type, str]
TypeList = Sequence[TypeHint]

FLOAT_TYPES = [float, "float"]
INTEGER_TYPES = [int, "int", "integer"]

DATETIME_TYPES = [datetime, "datetime"]
TIME_TYPES = [time, "time"]
TIMESTAMP_TYPES = ["timestamp"]
DATE_TYPES = [date, "date"]

STRING_TYPES = [str, "str", "string"]
LIST_TYPES = [list, List, "list"]
TUPLE_TYPES = [tuple, Tuple, "tuple"]

DICT_TYPES = [Dict, dict, "dict"]
RECORDS_TYPES = [Records, "records"]
DATAFRAME_TYPES = [DataFrame, "dataframe", "df", "pandas", "pd"]

TYPE_LIST = {
    float: FLOAT_TYPES,
    int: INTEGER_TYPES,
    datetime: DATETIME_TYPES,
    time: TIME_TYPES,
    date: DATE_TYPES,
    str: STRING_TYPES,
    list: LIST_TYPES,
    tuple: TUPLE_TYPES,
    dict: DICT_TYPES,
    DataFrame: DATAFRAME_TYPES,
}


def is_type(__type: TypeHint, __types: TypeList) -> bool:
    return (__type.lower() if isinstance(__type, str) else __type) in __types

def is_float_type(__type: TypeHint) -> bool:
    return is_type(__type, FLOAT_TYPES)

def is_int_type(__type: TypeHint) -> bool:
    return is_type(__type, INTEGER_TYPES)

def is_datetime_type(__type: TypeHint) -> bool:
    return is_type(__type, DATETIME_TYPES)

def is_time_type(__type: TypeHint) -> bool:
    return is_type(__type, TIME_TYPES)

def is_timestamp_type(__type: TypeHint) -> bool:
    return is_type(__type, TIMESTAMP_TYPES)

def is_date_type(__type: TypeHint) -> bool:
    return is_type(__type, DATE_TYPES)

def is_str_type(__type: TypeHint) -> bool:
    return is_type(__type, STRING_TYPES)

def is_list_type(__type: TypeHint) -> bool:
    return is_type(__type, LIST_TYPES)

def is_tuple_type(__type: TypeHint) -> bool:
    return is_type(__type, TUPLE_TYPES)

def is_dict_type(__type: TypeHint) -> bool:
    return is_type(__type, DICT_TYPES)

def is_records_type(__type: TypeHint) -> bool:
    return is_type(__type, RECORDS_TYPES)

def is_dataframe_type(__type: TypeHint) -> bool:
    return is_type(__type, DATAFRAME_TYPES)


###################################################################
######################### Array Type Check ########################
###################################################################

def is_array(__object) -> bool:
    return isinstance(__object, OBJECT_SEQUENCE)

def isin_instance(__object: Sequence, __type: Type, empty=False) -> bool:
    return (empty or __object) and any(map(lambda x: isinstance(x, __type), __object))

def type_isin_instance(__object: Sequence, __type: TypeHint, empty=False) -> bool:
    return (empty or __object) and any(map(lambda x: is_type(__type, TYPE_LIST[type(x)]), __object))

def allin_instance(__object: Sequence, __type: Type, empty=False) -> bool:
    return (empty or __object) and all(map(lambda x: isinstance(x, __type), __object))

def type_allin_instance(__object: Sequence, __type: TypeHint, empty=False) -> bool:
    return (empty or __object) and all(map(lambda x: is_type(__type, TYPE_LIST[type(x)]), __object))

def is_nested_in(__object, __type: Type, how="any", empty=False) -> bool:
    return isinstance(__object, OBJECT_SEQUENCE) and (
        isin_instance(__object, __type, empty) if how == "any" else allin_instance(__object, __type, empty))

def is_nested_map(__object, empty=False) -> bool:
    return isinstance(__object, Dict) and (empty or (bool(__object) and isin_instance(__object.values(), Dict)))

def is_float_array(__object, how="any", empty=False) -> bool:
    return is_nested_in(__object, float, how=how, empty=empty)

def is_int_array(__object, how="any", empty=False) -> bool:
    return is_nested_in(__object, int, how=how, empty=empty)

def is_datetime_array(__object, how="any", empty=False) -> bool:
    return is_nested_in(__object, datetime, how=how, empty=empty)

def is_time_array(__object, how="any", empty=False) -> bool:
    return is_nested_in(__object, time, how=how, empty=empty)

def is_date_array(__object, how="any", empty=False) -> bool:
    return is_nested_in(__object, date, how=how, empty=empty)

def is_str_array(__object, how="any", empty=False) -> bool:
    return is_nested_in(__object, str, how=how, empty=empty)

def is_2darray(__object, how="any", empty=False) -> bool:
    return is_nested_in(__object, OBJECT_SEQUENCE, how=how, empty=empty)

def is_records(__object, how="any", empty=True) -> bool:
    return is_nested_in(__object, Dict, how=how, empty=empty)

def is_dfarray(__object, how="any", empty=True) -> bool:
    return is_nested_in(__object, DataFrame, how=how, empty=empty)

def is_df(__object) -> bool:
    return isinstance(__object, DataFrame)
